# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The top level of the LiSE world model, the Character.

Based on NetworkX DiGraph objects with various additions and
conveniences.

A Character is a graph that follows rules. Its rules may be assigned
to run on only some portion of it: just edges (called Portals), just
nodes, or just nodes of the kind that have a location in another node
(called Places and Things, respectively). Each Character has a
``stat`` property that acts very much like a dictionary, in which you
can store game-relevant data for the rules to use.

You can designate some nodes in one Character as avatars of another,
and then assign a rule to run on all of a Character's avatars. This is
useful for the common case where someone in your game has a location
in the physical world (here, a Character, called 'physical') but also
has a behavior flowchart, or a skill tree, that isn't part of the
physical world. In that case the flowchart is the person's Character,
and their node in the physical world is an avatar of it.

"""

from abc import abstractmethod, ABC
from collections.abc import (Mapping, MutableMapping)
from itertools import chain
from types import MethodType
from typing import Type
from blinker import Signal

import networkx as nx
from .allegedb.cache import FuturistWindowDict, PickyDefaultDict
from .allegedb.graph import (DiGraph, GraphNodeMapping,
								DiGraphSuccessorsMapping,
								DiGraphPredecessorsMapping)
from .allegedb.wrap import MutableMappingUnwrapper

from .xcollections import CompositeDict
from .rule import RuleMapping
from .rule import RuleFollower as BaseRuleFollower
from .node import Node
from .thing import Thing
from .place import Place
from .portal import Portal
from .util import getatt, singleton_get, timer, AbstractEngine, AbstractCharacter
from .exc import WorldIntegrityError
from .query import StatusAlias


def grid_2d_8graph(m, n):
	"""Make a 2d graph that's connected 8 ways, with diagonals"""
	me = nx.Graph()
	nodes = me.nodes
	add_node = me.add_node
	add_edge = me.add_edge
	for i in range(m):
		for j in range(n):
			add_node((i, j))
			if i > 0:
				add_edge((i, j), (i - 1, j))
				if j > 0:
					add_edge((i, j), (i - 1, j - 1))
			if j > 0:
				add_edge((i, j), (i, j - 1))
			if (i - 1, j + 1) in nodes:
				add_edge((i, j), (i - 1, j + 1))
	return me


class CharRuleMapping(RuleMapping):
	"""Get rules by name, or make new ones by decorator

    You can access the rules in this either dictionary-style or as
    attributes. This is for convenience if you want to get at a rule's
    decorators, eg. to add an Action to the rule.

    Using this as a decorator will create a new rule, named for the
    decorated function, and using the decorated function as the
    initial Action.

    Using this like a dictionary will let you create new rules,
    appending them onto the underlying :class:`RuleBook`; replace one
    rule with another, where the new one will have the same index in
    the :class:`RuleBook` as the old one; and activate or deactivate
    rules. The name of a rule may be used in place of the actual rule,
    so long as the rule already exists.

    You can also set a rule active or inactive by setting it to
    ``True`` or ``False``, respectively. Inactive rules are still in
    the rulebook, but won't be followed.

    """

	def __init__(self, character, rulebook, booktyp):
		"""Initialize as usual for the ``rulebook``, mostly.

        My ``character`` property will be the one passed in, and my
        ``_table`` will be the ``booktyp`` with ``"_rules"`` appended.

        """
		super().__init__(rulebook.engine, rulebook)
		self.character = character
		self._table = booktyp + "_rules"


class RuleFollower(BaseRuleFollower):
	"""Mixin class. Has a rulebook, which you can get a RuleMapping into."""
	character: AbstractCharacter
	engine: AbstractEngine
	_book: str

	def _get_rule_mapping(self):
		return CharRuleMapping(self.character, self.rulebook, self._book)

	@abstractmethod
	def _get_rulebook_cache(self):
		pass

	def _get_rulebook_name(self):
		try:
			return self._get_rulebook_cache().retrieve(self.character.name,
														*self.engine._btt())
		except KeyError:
			return self.character.name, self._book

	def _set_rulebook_name(self, n):
		branch, turn, tick = self.engine._nbtt()
		self.engine.query._set_rulebook_on_character(self._book,
														self.character.name,
														branch, turn, tick, n)
		self._get_rulebook_cache().store(self.character.name, branch, turn,
											tick, n)

	def __contains__(self, k):
		return self.engine._active_rules_cache.contains_key(
			self._get_rulebook_name(), *self.engine._btt())


class FacadeEntity(MutableMapping, Signal, ABC):
	exists = True

	def __init__(self, mapping, _=None, **kwargs):
		super().__init__()
		self.facade = self.character = mapping.facade
		self._real = mapping
		self._patch = {
			k: v.unwrap() if hasattr(v, 'unwrap') else v
			for (k, v) in kwargs.items()
		}

	def __contains__(self, item):
		patch = self._patch
		return item in self._real or (item in patch
										and patch[item] is not None)

	def __iter__(self):
		seen = set()
		for k in self._real:
			if k not in self._patch:
				yield k
				seen.add(k)
		for k in self._patch:
			if (self._patch[k] is not None and k not in seen):
				yield k

	def __len__(self):
		n = 0
		for k in self:
			n += 1
		return n

	def __getitem__(self, k):
		if k in self._patch:
			if self._patch[k] is None:
				raise KeyError("{} has been masked.".format(k))
			return self._patch[k]
		ret = self._real[k]
		if hasattr(ret, 'unwrap'):  # a wrapped mutable object from the
			# allegedb.wrap module
			ret = ret.unwrap()
			self._patch[k] = ret  # changes will be reflected in the
		# facade but not the original
		return ret

	def __setitem__(self, k, v):
		if k == 'name':
			raise TypeError("Can't change names")
		if hasattr(v, 'unwrap'):
			v = v.unwrap()
		self._patch[k] = v
		self.send(self, key=k, val=v)

	def __delitem__(self, k):
		self._patch[k] = None
		self.send(self, key=k, val=None)


class FacadeNode(FacadeEntity, ABC):

	@property
	def name(self):
		return self['name']

	@property
	def portal(self):
		return self.facade.portal[self['name']]

	def contents(self):
		for thing in self.facade.thing.values():
			# it seems like redundant FacadeNode are being created sometimes
			if thing['location'] == self.name:
				yield thing


class FacadePlace(FacadeNode):
	"""Lightweight analogue of Place for Facade use."""

	def __init__(self, mapping, real_or_name, **kwargs):
		super().__init__(mapping, real_or_name, **kwargs)
		if isinstance(real_or_name, Place) or isinstance(
			real_or_name, FacadePlace):
			self._real = real_or_name
		else:
			self._real = {'name': real_or_name}

	def add_thing(self, name):
		self.facade.add_thing(name, self.name)

	def new_thing(self, name):
		return self.facade.new_thing(name, self.name)


class FacadeThing(FacadeNode):

	def __init__(self, mapping, real_or_name, **kwargs):
		location = kwargs.pop('location', None)
		super().__init__(mapping, real_or_name, **kwargs)
		if location is None and not (isinstance(real_or_name, Thing) or
										isinstance(real_or_name, FacadeThing)):
			raise TypeError(
				"FacadeThing needs to wrap a real Thing or another "
				"FacadeThing, or have a location of its own.")
		self._real = {
			'name': real_or_name.name
			if hasattr(real_or_name, 'name') else real_or_name,
			'location': location
		}

	@property
	def location(self):
		return self.facade.node[self['location']]

	@location.setter
	def location(self, v):
		if isinstance(v, (FacadePlace, FacadeThing)):
			v = v.name
		if v not in self.facade.node:
			raise KeyError("Location {} not present".format(v))
		self['location'] = v


class FacadePortal(FacadeEntity):
	"""Lightweight analogue of Portal for Facade use."""

	def __init__(self, mapping, other, **kwargs):
		super().__init__(mapping, other, **kwargs)
		if hasattr(mapping, 'orig'):
			self.orig = mapping.orig
			self.dest = other
		else:
			self.dest = mapping.dest
			self.orig = other
		try:
			self._real = self.facade.character.portal[self.orig][self.dest]
		except (KeyError, AttributeError):
			self._real = {}

	def __getitem__(self, item):
		if item == 'origin':
			return self.orig
		if item == 'destination':
			return self.dest
		return super().__getitem__(item)

	def __setitem__(self, k, v):
		if k in ('origin', 'destination'):
			raise TypeError("Portals have fixed origin and destination")
		super().__setitem__(k, v)

	@property
	def origin(self):
		return self.facade.node[self.orig]

	@property
	def destination(self):
		return self.facade.node[self.dest]


class FacadeEntityMapping(MutableMappingUnwrapper, Signal, ABC):
	"""Mapping that contains entities in a Facade.

    All the entities are of the same type, ``facadecls``, possibly
    being distorted views of entities of the type ``innercls``.

    """
	facadecls: Type[FacadeEntity]

	@abstractmethod
	def _get_inner_map(self):
		raise NotImplementedError("Missing _get_inner_map")

	def _make(self, k, v):
		kwargs = dict(v)
		for badkey in ('character', 'engine', 'name'):
			if badkey in kwargs:
				del kwargs[badkey]
		return self.facadecls(self, k, **kwargs)

	engine = getatt('facade.engine')

	def __init__(self, facade, _=None):
		"""Store the facade."""
		super().__init__()
		self.facade = facade
		self._patch = {}

	def __contains__(self, k):
		if k in self._patch:
			return self._patch[k] is not None
		return k in self._get_inner_map()

	def __iter__(self):
		seen = set()
		for k in self._patch:
			if k not in seen and self._patch[k] is not None:
				yield k
			seen.add(k)
		for k in self._get_inner_map():
			if k not in seen:
				yield k

	def __len__(self):
		n = 0
		for k in self:
			n += 1
		return n

	def __getitem__(self, k):
		if k not in self:
			raise KeyError
		if k not in self._patch:
			self._patch[k] = self._make(k, self._get_inner_map()[k])
		ret = self._patch[k]
		if ret is None:
			raise KeyError
		if type(ret) is not self.facadecls:
			ret = self._patch[k] = self._make(k, ret)
		return ret

	def __setitem__(self, k, v):
		if not isinstance(v, self.facadecls):
			v = self._make(k, v)
		self._patch[k] = v
		self.send(self, key=k, val=v)

	def __delitem__(self, k):
		if k not in self:
			raise KeyError("{} not present".format(k))
		self._patch[k] = None
		self.send(self, key=k, val=None)


class FacadePortalSuccessors(FacadeEntityMapping):
	facadecls = FacadePortal
	innercls = Portal

	def __init__(self, facade, origname):
		super().__init__(facade, origname)
		self.orig = origname

	def _make(self, k, v):
		return self.facadecls(self, k, **v)

	def _get_inner_map(self):
		try:
			return self.facade.character.portal[self.orig]
		except AttributeError:
			return {}


class FacadePortalPredecessors(FacadeEntityMapping):
	facadecls = FacadePortal
	innercls = Portal

	def __init__(self, facade, destname):
		super().__init__(facade, destname)
		self.dest = destname

	def _make(self, k, v):
		return self.facadecls(self.facade.portal[k], v)

	def _get_inner_map(self):
		try:
			return self.facade.character.preportal[self.dest]
		except AttributeError:
			return {}


class FacadePortalMapping(FacadeEntityMapping, ABC):
	cls: Type[FacadeEntityMapping]

	def __getitem__(self, node):
		if node not in self:
			raise KeyError("No such node: {}".format(node))
		if node not in self._patch:
			self._patch[node] = self.cls(self.facade, node)
		ret = self._patch[node]
		if ret is None:
			raise KeyError("masked")
		if type(ret) is not self.cls:
			nuret = self.cls(self.facade, node)
			if type(ret) is dict:
				nuret._patch = ret
			else:
				nuret.update(ret)
			ret = nuret
		return ret


class Facade(AbstractCharacter, nx.DiGraph):
	engine = getatt('character.engine')
	db = getatt('character.engine')

	def __getstate__(self):
		ports = {}
		for o in self.portal:
			if o not in ports:
				ports[o] = {}
			for d in self.portal[o]:
				ports[o][d] = dict(self.portal[o][d])
		things = {k: dict(v) for (k, v) in self.thing.items()}
		places = {k: dict(v) for (k, v) in self.place.items()}
		stats = {
			k: v.unwrap() if hasattr(v, 'unwrap') else v
			for (k, v) in self.graph.items()
		}
		return things, places, ports, stats

	def __setstate__(self, state):
		self.character = None
		self.graph = self.StatMapping(self)
		(self.thing._patch, self.place._patch, self.portal._patch,
			self.graph._patch) = state

	def add_places_from(self, seq, **attrs):
		for place in seq:
			self.add_place(place, **attrs)

	def add_things_from(self, seq, **attrs):
		for thing in seq:
			self.add_thing(thing, **attrs)

	def thing2place(self, name):
		self.place[name] = self.thing.pop(name)

	def place2thing(self, name, location):
		it = self.place.pop(name)
		it['location'] = location
		self.thing[name] = it

	def add_portals_from(self, seq, **attrs):
		for it in seq:
			self.add_portal(*it, **attrs)

	def remove_unit(self, a, b=None):
		raise NotImplementedError("Facades don't have units")

	def add_place(self, name, **kwargs):
		self.place[name] = kwargs

	def add_node(self, name, **kwargs):
		self.place[name] = kwargs

	def remove_node(self, node):
		if node in self.thing:
			del self.thing[node]
		else:
			del self.place[node]

	def remove_place(self, place):
		del self.place[place]

	def remove_thing(self, thing):
		del self.thing[thing]

	def add_thing(self, name, location, **kwargs):
		kwargs['location'] = location
		self.thing[name] = kwargs

	def add_portal(self, orig, dest, symmetrical=False, **kwargs):
		self.portal[orig][dest] = kwargs
		if symmetrical:
			mirror = dict(kwargs)
			mirror['is_mirror'] = True
			self.portal[dest][orig] = mirror

	def remove_portal(self, origin, destination):
		del self.portal[origin][destination]

	def add_edge(self, orig, dest, **kwargs):
		self.add_portal(orig, dest, **kwargs)

	def add_unit(self, a, b=None):
		raise NotImplementedError("Facades don't have units")

	def __init__(self, character=None):
		"""Store the character."""
		super().__init__()
		self.character = character
		self.graph = self.StatMapping(self)

	class ThingMapping(FacadeEntityMapping):
		facadecls = FacadeThing
		innercls = Thing

		def _get_inner_map(self):
			try:
				return self.facade.character.thing
			except AttributeError:
				return {}

		def patch(self, d: dict):
			places = d.keys() & self.facade.place.keys()
			if places:
				raise KeyError(
					f"Tried to patch places on thing mapping: {places}")
			self.facade.node.patch(d)

	class PlaceMapping(FacadeEntityMapping):
		facadecls = FacadePlace
		innercls = Place

		def _get_inner_map(self):
			try:
				return self.facade.character._node
			except AttributeError:
				return {}

		def patch(self, d: dict):
			things = d.keys() & self.facade.thing.keys()
			if things:
				raise KeyError(
					f"Tried to patch things on place mapping: {things}")
			self.facade.node.patch(d)

	def ThingPlaceMapping(self, *args):
		return CompositeDict(self.place, self.thing)

	class PortalSuccessorsMapping(FacadePortalMapping):
		cls = FacadePortalSuccessors

		def __contains__(self, item):
			return item in self.facade.node

		def _get_inner_map(self):
			try:
				return self.facade.character._adj
			except AttributeError:
				return {}

	class PortalPredecessorsMapping(FacadePortalMapping):
		cls = FacadePortalPredecessors

		def __contains__(self, item):
			return item in self.facade._node

		def _get_inner_map(self):
			try:
				return self.facade.character.pred
			except AttributeError:
				return {}

	class StatMapping(MutableMappingUnwrapper, Signal):

		def __init__(self, facade):
			super().__init__()
			self.facade = facade
			self._patch = {}

		def __iter__(self):
			seen = set()
			if hasattr(self.facade.character, 'graph'):
				for k in self.facade.character.graph:
					if k not in self._patch:
						yield k
						seen.add(k)
			for (k, v) in self._patch.items():
				if k not in seen and v is not None:
					yield k

		def __len__(self):
			n = 0
			for k in self:
				n += 1
			return n

		def __contains__(self, k):
			if hasattr(self.facade.character,
						'graph') and k in self.facade.character.graph:
				return True
			return k in self._patch and self._patch[k] is not None

		def __getitem__(self, k):
			if k not in self._patch and hasattr(self.facade.character,
												'graph'):
				ret = self.facade.character.graph[k]
				if not hasattr(ret, 'unwrap'):
					return ret
				self._patch[k] = ret.unwrap()
			if self._patch[k] is None:
				return KeyError
			return self._patch[k]

		def __setitem__(self, k, v):
			self._patch[k] = v
			self.send(self, key=k, val=v)

		def __delitem__(self, k):
			self._patch[k] = None
			self.send(self, key=k, val=None)


class Character(DiGraph, AbstractCharacter, RuleFollower):
	"""A digraph that follows game rules and has a containment hierarchy

    Nodes in a Character are subcategorized into Things and
    Places. Things have locations, and those locations may be Places
    or other Things. To get at those, use the `thing` and `place`
    mappings -- but in situations where the distinction does not matter,
    you may simply address the Character as a mapping, as in NetworkX.

    Characters may have units in other Characters. These are just
    nodes. You can apply rules to a Character's units, and thus to
    any collection of nodes you want, perhaps in many different
    Characters. The `unit` attribute handles this. It is a mapping,
    keyed by the other Character's name, then by the name of the node
    that is this Character's unit. In the common case where a
    Character has exactly one unit, it may be retrieved as
    `unit.only`. When it has more than one unit, but only has
    any units in a single other Character, you can get the mapping
    of units in that Character as `unit.node`. Add units with the
    `add_unit` method and remove them with `del_unit`.

    You can assign rules to Characters with their `rule` attribute,
    typically using it as a decorator (see the documentation for
    the `rule` module). You can do the same to some of Character's
    attributes:

    * `thing.rule` to make a rule run on all Things in this Character
      every turn
    * `place.rule` to make a rule run on all Places in this Character
      every turn
    * `node.rule` to make a rule run on all Things and Places in this
      Character every turn
    * `unit.rule` to make a rule run on all the units this
      Character has every turn, regardless of what Character the
      unit is in
    * `adj.rule` to make a rule run on all the edges this Character
      has every turn

    """
	_book = "character"
	remove_portal = getatt('remove_edge')

	@property
	def character(self):
		return self

	def _get_rulebook_cache(self):
		return self.engine._characters_rulebooks_cache

	def __repr__(self):
		return "{}.character[{}]".format(repr(self.engine), repr(self.name))

	def __init__(self, engine, name, *, init_rulebooks=True):
		super().__init__(engine, name)
		self._avatars_cache = PickyDefaultDict(FuturistWindowDict)
		if not init_rulebooks:
			return
		cachemap = {
			'character': engine._characters_rulebooks_cache,
			'unit': engine._units_rulebooks_cache,
			'character_thing': engine._characters_things_rulebooks_cache,
			'character_place': engine._characters_places_rulebooks_cache,
			'character_portal': engine._characters_portals_rulebooks_cache
		}
		branch, turn, tick = engine._btt()
		for rulebook, cache in cachemap.items():
			rulebook_name = (name, rulebook)
			engine.query._set_rulebook_on_character(rulebook, name, branch,
													turn, tick, rulebook_name)
			cache.store((name, rulebook), branch, turn, tick, rulebook_name)

	class ThingMapping(MutableMappingUnwrapper, RuleFollower, Signal):
		""":class:`Thing` objects that are in a :class:`Character`"""
		_book = "character_thing"

		engine = getatt('character.engine')
		name = getatt('character.name')

		def _get_rulebook_cache(self):
			return self.engine._characters_things_rulebooks_cache

		def __init__(self, character):
			"""Store the character and initialize cache."""
			super().__init__()
			self.character = character

		def __iter__(self):
			cache = self.engine._things_cache
			char = self.name
			branch, turn, tick = self.engine._btt()
			for key in cache.iter_keys(char, branch, turn, tick):
				try:
					if cache.retrieve(char, key, branch, turn,
										tick) is not None:
						yield key
				except KeyError:
					continue

		def __contains__(self, thing):
			branch, turn, tick = self.engine._btt()
			args = self.character.name, thing, branch, turn, tick
			cache = self.engine._things_cache
			return cache.contains_key(*args)

		def __len__(self):
			return self.engine._things_cache.count_keys(
				self.character.name, *self.engine._btt())

		def __getitem__(self, thing):
			if thing not in self:
				raise KeyError("No such thing: {}".format(thing))
			return self._make_thing(thing)

		def _make_thing(self, thing, val=None):
			cache = self.engine._node_objs
			if isinstance(val, Thing):
				th = cache[self.name, thing] = val
			elif (self.name, thing) in cache:
				th = cache[(self.name, thing)]
				if type(th) is not Thing:
					th = cache[self.name, thing] = Thing(self.character, thing)
			else:
				th = cache[(self.name, thing)] = Thing(self.character, thing)
			return th

		def __setitem__(self, thing, val):
			if not isinstance(val, Mapping):
				raise TypeError('Things are made from Mappings')
			if 'location' not in val:
				raise ValueError('Thing needs location')
			created = thing not in self
			self.engine._exist_node(self.character.name, thing)
			self.engine._set_thing_loc(self.character.name, thing,
										val['location'])
			th = self._make_thing(thing, val)
			th.clear()
			th.update(val)
			if created:
				self.send(self, thing_name=thing, exists=True)

		def __delitem__(self, thing):
			self[thing].delete()
			self.send(self, thing_name=thing, exists=False)

		def __repr__(self):
			return "{}.character[{}].thing".format(repr(self.engine),
													repr(self.name))

	class PlaceMapping(MutableMappingUnwrapper, RuleFollower, Signal):
		""":class:`Place` objects that are in a :class:`Character`"""
		_book = "character_place"

		def _get_rulebook_cache(self):
			return self.engine._characters_places_rulebooks_cache

		def update(self, __m, **kwargs) -> None:
			engine = self.engine
			store_node = engine._nodes_cache.store
			store_node_val = engine._node_val_cache.store
			iter_node_keys = engine._node_val_cache.iter_keys
			exist_node = engine.query.exist_node
			node_val_set = engine.query.node_val_set
			branch, turn, start_tick = engine._btt()
			tick = start_tick + 1
			charn = self.character.name
			planning = engine._planning
			forward = engine._forward
			with timer("seconds spent updating PlaceMapping"):
				for node, val in chain(__m.items(), kwargs.items()):
					if val is None:
						for key in iter_node_keys(charn,
													node,
													branch,
													turn,
													start_tick,
													forward=forward):
							store_node_val(charn,
											node,
											key,
											branch,
											turn,
											tick,
											None,
											planning=planning,
											forward=forward,
											loading=True)
							node_val_set(charn, node, key, branch, turn, tick,
											None)
							tick += 1
						store_node(charn,
									node,
									branch,
									turn,
									tick,
									False,
									planning=planning,
									forward=forward,
									loading=True)
						exist_node(charn, node, branch, turn, tick, False)
						tick += 1
					else:
						store_node(charn,
									node,
									branch,
									turn,
									tick,
									True,
									planning=planning,
									forward=forward,
									loading=True)
						exist_node(charn, node, branch, turn, tick, True)
						tick += 1
						for k, v in val.items():
							store_node_val(charn,
											node,
											k,
											branch,
											turn,
											tick,
											v,
											planning=planning,
											forward=forward,
											loading=True)
							exist_node(charn, node, k, branch, turn, tick, v)
							tick += 1
			engine.tick = tick

		def __init__(self, character):
			"""Store the character."""
			super().__init__()
			self.character = character
			self.engine = engine = character.engine
			charn = character.name
			nodes_cache = engine._nodes_cache
			things_cache = engine._things_cache
			iter_nodes = nodes_cache.iter_entities
			nodes_contains = nodes_cache.contains_entity
			things_contains = things_cache.contains_entity
			btt = engine._btt
			self._iter_stuff = (iter_nodes, things_contains, charn, btt)
			self._len_stuff = (nodes_cache.count_entities,
								things_cache.count_entities, charn, btt)
			self._contains_stuff = (nodes_contains, things_contains, charn,
									btt)
			self._get_stuff = self._contains_stuff + (engine._node_objs,
														character)
			self._set_stuff = (engine._node_exists, engine._exist_node,
								engine._get_node, charn, character)

		def __iter__(self):
			iter_nodes, things_contains, charn, btt = self._iter_stuff
			branch, turn, tick = btt()
			for node in iter_nodes(charn, branch, turn, tick):
				if not things_contains(charn, node, branch, turn, tick):
					yield node

		def __len__(self):
			count_nodes, count_things, charn, btt = self._len_stuff
			branch, turn, tick = btt()
			return count_nodes(charn, branch, turn, tick) - count_things(
				charn, branch, turn, tick)

		def __contains__(self, place):
			# TODO: maybe a special cache just for places and not just
			# nodes in general
			nodes_contains, things_contains, charn, btt = self._contains_stuff
			branch, turn, tick = btt()
			return (nodes_contains(charn, place, branch, turn, tick)
					and not things_contains(charn, place, branch, turn, tick))

		def __getitem__(self, place):
			nodes_contains, things_contains, charn, btt, cache, character = self._get_stuff
			branch, turn, tick = btt()
			if not nodes_contains(charn, place, branch, turn,
									tick) or things_contains(
										charn, place, branch, turn, tick):
				raise KeyError("No such place: {}".format(place))
			if (charn, place) not in cache or not isinstance(
				cache[(charn, place)], Place):
				ret = cache[(charn, place)] = Place(character, place)
				return ret
			return cache[(charn, place)]

		def __setitem__(self, place, v):
			node_exists, exist_node, get_node, charn, character = self._set_stuff
			exist_node(charn, place, True)
			pl = get_node(character, place)
			if not isinstance(pl, Place):
				raise KeyError("{} is a {}, not a place".format(
					place,
					type(pl).__name__))
			pl.update(v)
			self.send(self, key=place, val=v)

		def __delitem__(self, place):
			self[place].delete()

		def __repr__(self):
			return "{}.character[{}].place".format(repr(self.character.engine),
													repr(self.character.name))

	class ThingPlaceMapping(GraphNodeMapping, Signal):
		"""GraphNodeMapping but for Place and Thing"""
		_book = "character_node"

		character = getatt('graph')
		engine = getatt('db')
		name = getatt('character.name')

		def __init__(self, character):
			"""Store the character."""
			super().__init__(character)
			Signal.__init__(self)
			engine = character.engine
			charn = character.name
			self._contains_stuff = contains_stuff = (engine._node_exists,
														charn)
			self._getitem_stuff = contains_stuff + (engine._get_node,
													character)
			self._delitem_stuff = contains_stuff + (
				engine._is_thing, character.thing, character.place)
			self._placemap = character.place

		def __contains__(self, k):
			node_exists, charn = self._contains_stuff
			return node_exists(charn, k)

		def __getitem__(self, k):
			node_exists, charn, get_node, character = self._getitem_stuff
			if not node_exists(charn, k):
				raise KeyError
			return get_node(character, k)

		def __setitem__(self, k, v):
			self._placemap[k] = v

		def __delitem__(self, k):
			node_exists, charn, is_thing, thingmap, placemap = self._delitem_stuff
			if not node_exists(charn, k):
				raise KeyError
			if is_thing(charn, k):
				del thingmap[k]
			else:
				del placemap[k]

	node_map_cls = ThingPlaceMapping

	class PortalSuccessorsMapping(DiGraphSuccessorsMapping, RuleFollower):
		"""Mapping of nodes that have at least one outgoing edge.

        Maps them to another mapping, keyed by the destination nodes,
        which maps to Portal objects.

        """
		_book = "character_portal"

		character = getatt('graph')
		engine = getatt('graph.engine')
		upd_succs_time = 0

		def __init__(self, graph):
			super().__init__(graph)
			engine = graph.engine
			charn = graph.name
			self._cporh = engine._characters_portals_rulebooks_cache
			self._getitem_stuff = (engine._node_exists, charn, self._cache)
			self._setitem_stuff = (self._cache, self.Successors)

		def _get_rulebook_cache(self):
			return self._cporh

		def __getitem__(self, orig):
			node_exists, charn, cache = self._getitem_stuff
			if node_exists(charn, orig):
				if orig not in cache:
					cache[orig] = self.Successors(self, orig)
				return cache[orig]
			raise KeyError("No such node")

		def __delitem__(self, orig):
			super().__delitem__(orig)
			self.send(self, key=orig, val=None)

		def update(self, other, **kwargs):
			"""Recursively update the stats of all portals

            Input should be a dictionary of dictionaries of dictionaries
            --just like networkx ``DiGraph._edge``.

            This will create portals as needed, but will only delete
            them if you set their value to ``None``. Likewise, stats
            not specified in the input will be left untouched, if they
            are already present, but you can set them to ``None`` to
            delete them.

            """
			engine = self.engine
			planning = engine._planning
			forward = engine._forward
			branch, turn, start_tick = engine._btt()
			exist_edge = engine.query.exist_edge
			edge_val_set = engine.query.edge_val_set
			store_edge = engine._edges_cache.store
			store_edge_val = engine._edge_val_cache.store
			iter_edge_keys = engine._edge_val_cache.iter_entity_keys
			charn = self.character.name
			tick = start_tick + 1
			with timer("seconds spent updating PortalSuccessorsMapping"):
				for orig, dests in chain(other.items(), kwargs.items()):
					for dest, kvs in dests.items():
						if kvs is None:
							for k in iter_edge_keys(charn,
													orig,
													dest,
													0,
													branch,
													turn,
													start_tick,
													forward=forward):
								store_edge_val(charn,
												orig,
												dest,
												0,
												k,
												branch,
												turn,
												tick,
												None,
												planning=planning,
												forward=forward,
												loading=True)
								edge_val_set(charn, orig, dest, 0, k, branch,
												turn, tick, None)
								tick += 1
							store_edge(charn,
										orig,
										dest,
										0,
										branch,
										turn,
										tick,
										False,
										planning=planning,
										forward=forward,
										loading=True)
							exist_edge(charn, orig, dest, 0, branch, turn,
										tick, False)
							tick += 1
						else:
							store_edge(charn,
										orig,
										dest,
										0,
										branch,
										turn,
										tick,
										True,
										planning=planning,
										forward=forward,
										loading=True)
							exist_edge(charn, orig, dest, 0, branch, turn,
										tick, True)
							tick += 1
							for k, v in kvs.items():
								store_edge_val(charn,
												orig,
												dest,
												0,
												k,
												branch,
												turn,
												tick,
												v,
												planning=planning,
												forward=forward,
												loading=True)
								edge_val_set(charn, orig, dest, 0, k, branch,
												turn, tick, v)
								tick += 1
			engine.tick = tick

		class Successors(DiGraphSuccessorsMapping.Successors):
			"""Mapping for possible destinations from some node."""

			engine = getatt('graph.engine')

			@staticmethod
			def send(self, **kwargs):
				"""Call all listeners to ``dest`` and to my ``orig``."""
				super().send(self, **kwargs)
				self.container.send(self, **kwargs)

			def __init__(self, container, orig):
				super().__init__(container, orig)
				graph = self.graph
				engine = graph.engine
				self._getitem_stuff = (engine._get_edge, graph, orig)
				self._setitem_stuff = (engine._edge_exists, engine._exist_edge,
										graph.name, orig, engine._get_edge,
										graph, engine.query.edge_val_set,
										engine._edge_val_cache.store,
										engine._nbtt)

			def __getitem__(self, dest):
				get_edge, graph, orig = self._getitem_stuff
				if dest in self:
					return get_edge(graph, orig, dest, 0)
				raise KeyError("No such portal: {}->{}".format(orig, dest))

			def __setitem__(self, dest, value):
				if value is None:
					del self[dest]
					return
				(edge_exists, exist_edge, charn, orig, get_edge, graph,
					db_edge_val_set, edge_val_cache_store,
					nbtt) = self._setitem_stuff
				exist_edge(charn, orig, dest)
				for k, v in value.items():
					branch, turn, tick = nbtt()
					db_edge_val_set(charn, orig, dest, 0, k, branch, turn,
									tick, v)
					edge_val_cache_store(charn, orig, dest, 0, k, branch, turn,
											tick, v)
				self.send(self, key=dest, val=value)

			def __delitem__(self, dest):
				if dest not in self:
					raise KeyError("No portal to {}".format(dest))
				self[dest].delete()

			def update(self, other, **kwargs):
				charn = self.graph.name
				orig = self.orig
				engine = self.engine
				store_edge = engine._edges_cache.store
				exist_edge = engine.query.exist_edge
				store_edge_val = engine._edge_val_cache.store
				set_edge_val = engine.query.edge_val_set
				iter_edge_keys = engine._edge_val_cache.iter_entity_keys
				planning = engine._planning
				forward = engine._forward
				branch, turn, start_tick = engine._btt()
				tick = start_tick + 1
				for dest, val in chain(other.items(), kwargs.items()):
					if val is None:
						for k in iter_edge_keys(charn, orig, dest, 0, branch,
												turn, start_tick):
							store_edge_val(charn,
											orig,
											dest,
											0,
											k,
											branch,
											turn,
											tick,
											None,
											planning=planning,
											forward=forward,
											loading=True)
							set_edge_val(charn, orig, dest, 0, k, branch, turn,
											tick, None)
							tick += 1
						store_edge(charn,
									orig,
									dest,
									0,
									branch,
									turn,
									tick,
									None,
									planning=planning,
									forward=forward,
									loading=True)
						exist_edge(charn, orig, dest, 0, branch, turn, tick,
									None)
						tick += 1

	adj_cls = PortalSuccessorsMapping

	class PortalPredecessorsMapping(DiGraphPredecessorsMapping, RuleFollower):
		"""Mapping of nodes that have at least one incoming edge.

        Maps to another mapping keyed by the origin nodes, which maps to
        Portal objects.

        """
		_book = "character_portal"

		def __init__(self, graph):
			super().__init__(graph)
			self._cporc = graph.engine._characters_portals_rulebooks_cache

		def _get_rulebook_cache(self):
			return self._cporc

		class Predecessors(DiGraphPredecessorsMapping.Predecessors):
			"""Mapping of possible origins from some destination."""

			def __init__(self, container, dest):
				super().__init__(container, dest)
				graph = self.graph
				self._setitem_stuff = (graph, graph.name, dest,
										self.db._edge_objs)

			def __setitem__(self, orig, value):
				graph, graph_name, dest, portal_objs = self._setitem_stuff
				key = (graph_name, orig, dest)
				if key not in portal_objs:
					portal_objs[key] = Portal(graph, orig, dest)
				p = portal_objs[key]
				p.clear()
				p.update(value)
				p.engine._exist_edge(graph_name, dest, orig)

	pred_cls = PortalPredecessorsMapping

	class UnitGraphMapping(Mapping, RuleFollower):
		"""A mapping of other characters in which one has a unit.

        Maps to a mapping of the units themselves, unless there's
        only one other character you have units in, in which case
        this maps to those.

        """
		_book = "unit"

		engine = getatt('character.engine')
		name = getatt('character.name')

		def _get_rulebook_cache(self):
			return self._avrc

		def __init__(self, char):
			"""Remember my character."""
			self.character = char
			self._char_av_cache = {}
			engine = char.engine
			self._avrc = engine._units_rulebooks_cache
			self._add_av = char.add_unit
			avcache = engine._unitness_cache
			get_char_graphs = avcache.get_char_graphs
			charn = char.name
			btt = engine._btt
			self._iter_stuff = (get_char_graphs, charn, btt)
			self._node_stuff = (self._get_char_av_cache,
								avcache.get_char_only_graph, charn, btt)
			self._only_stuff = (avcache.get_char_only_av, charn, btt,
								engine._get_node, engine.character)

		def __call__(self, av):
			"""Add the unit

            It must be an instance of Place or Thing.

            """
			if av.__class__ not in (Place, Thing):
				raise TypeError("Only Things and Places may be units")
			self._add_av(av.name, av.character.name)

		def __iter__(self):
			"""Iterate over graphs with unit nodes in them"""
			get_char_graphs, charn, btt = self._iter_stuff
			return iter(get_char_graphs(charn, *btt()))

		def __contains__(self, k):
			get_char_graphs, charn, btt = self._iter_stuff
			return k in get_char_graphs(charn, *btt())

		def __len__(self):
			"""Number of graphs in which I have a unit."""
			get_char_graphs, charn, btt = self._iter_stuff
			return len(get_char_graphs(charn, *btt()))

		def _get_char_av_cache(self, g):
			if g not in self:
				raise KeyError
			if g not in self._char_av_cache:
				self._char_av_cache[g] = self.CharacterUnitMapping(self, g)
			return self._char_av_cache[g]

		def __getitem__(self, g):
			return self._get_char_av_cache(g)

		@property
		def node(self):
			"""If I have units in only one graph, return a map of them

            Otherwise, raise AttributeError.

            """
			get_char_av_cache: MethodType
			get_char_av_cache, get_char_only_graph, charn, btt = self._node_stuff
			try:
				return get_char_av_cache(get_char_only_graph(charn, *btt()))
			except KeyError:
				raise AttributeError(
					"I have no unit, or I have units in many graphs")

		@property
		def only(self):
			"""If I have only one unit, this is it

            Otherwise, raise AttributeError.

            """
			get_char_only_av, charn, btt, get_node, charmap = self._only_stuff
			try:
				charn, noden = get_char_only_av(charn, *btt())
				return get_node(charmap[charn], noden)
			except KeyError:
				raise AttributeError("I have no unit, or more than one unit")

		class CharacterUnitMapping(Mapping):
			"""Mapping of units of one Character in another Character."""

			def __init__(self, outer, graphn):
				"""Store this character and the name of the other one"""
				self.character = character = outer.character
				self.engine = engine = outer.engine
				self.name = name = outer.name
				self.graph = graphn
				avcache = engine._unitness_cache
				btt = engine._btt
				self._iter_stuff = iter_stuff = (avcache.get_char_graph_avs,
													name, graphn, btt)
				get_node = engine._get_node
				self._getitem_stuff = iter_stuff + (get_node, graphn,
													engine.character)
				self._only_stuff = (get_node, engine.character, graphn)

			def __iter__(self):
				"""Iterate over names of unit nodes"""
				get_char_graph_avs, name, graphn, btt = self._iter_stuff
				return iter(get_char_graph_avs(name, graphn, *btt()))

			def __contains__(self, av):
				get_char_graph_avs, name, graphn, btt = self._iter_stuff
				return av in get_char_graph_avs(name, graphn, *btt())

			def __len__(self):
				"""Number of units of this character in that graph"""
				get_char_graph_avs, name, graphn, btt = self._iter_stuff
				return len(get_char_graph_avs(name, graphn, *btt()))

			def __getitem__(self, av):
				(get_char_graph_avs, name, graphn, btt, get_node, graphn,
					charmap) = self._getitem_stuff
				if av in get_char_graph_avs(name, graphn, *btt()):
					return get_node(charmap[graphn], av)
				raise KeyError("No unit: {}".format(av))

			@property
			def only(self):
				"""If I have only one unit, return it; else error"""
				mykey = singleton_get(self.keys())
				if mykey is None:
					raise AttributeError("No unit, or more than one")
				get_node, charmap, graphn = self._only_stuff
				return get_node(charmap[graphn], mykey)

			def __repr__(self):
				return "{}.character[{}].unit".format(repr(self.engine),
														repr(self.name))

	def facade(self):
		return Facade(self)

	def add_place(self, node_for_adding, **attr):
		self.add_node(node_for_adding, **attr)

	def add_places_from(self, seq, **attrs):
		"""Take a series of place names and add the lot."""
		super().add_nodes_from(seq, **attrs)

	def remove_place(self, place):
		if place in self.place:
			self.remove_node(place)
		raise KeyError("No such place: {}".format(place))

	def remove_thing(self, thing):
		if thing in self.thing:
			self.remove_node(thing)
		raise KeyError("No such thing: {}".format(thing))

	def add_thing(self, name, location, **kwargs):
		"""Make a new Thing and set its location"""
		if name in self.thing:
			raise WorldIntegrityError(
				"Already have a Thing named {}".format(name))
		self.add_node(name, **kwargs)
		if isinstance(location, Node):
			location = location.name
		self.place2thing(
			name,
			location,
		)

	def add_things_from(self, seq, **attrs):
		for tup in seq:
			name = tup[0]
			location = tup[1]
			kwargs = tup[2] if len(tup) > 2 else attrs
			self.add_thing(name, location, **kwargs)

	def place2thing(self, name, location):
		"""Turn a Place into a Thing with the given location.
        
        It will keep all its attached Portals.

        """
		self.engine._set_thing_loc(self.name, name, location)
		if (self.name, name) in self.engine._node_objs:
			obj = self.engine._node_objs[self.name, name]
			thing = Thing(self, name, clobber=True)
			for port in obj.portals():
				port.origin = thing
			for port in obj.preportals():
				port.destination = thing
			self.engine._node_objs[self.name, name] = thing

	def thing2place(self, name):
		"""Unset a Thing's location, and thus turn it into a Place."""
		self.engine._set_thing_loc(self.name, name, None)
		if (self.name, name) in self.engine._node_objs:
			thing = self.engine._node_objs[self.name, name]
			place = Place(self, name)
			for port in thing.portals():
				port.origin = place
			for port in thing.preportals():
				port.destination = place
			self.engine._node_objs[self.name, name] = place

	def add_portal(self, origin, destination, symmetrical=False, **kwargs):
		"""Connect the origin to the destination with a :class:`Portal`.

        Keyword arguments are the :class:`Portal`'s
        attributes. Exception: if keyword ``symmetrical`` == ``True``,
        a mirror-:class:`Portal` will be placed in the opposite
        direction between the same nodes. It will always appear to
        have the placed :class:`Portal`'s stats, and any change to the
        mirror :class:`Portal`'s stats will affect the placed
        :class:`Portal`.

        """
		if isinstance(origin, Node):
			origin = origin.name
		if isinstance(destination, Node):
			destination = destination.name
		super().add_edge(origin, destination, **kwargs)
		if symmetrical:
			self.add_portal(destination, origin, is_mirror=True)

	def new_portal(self, origin, destination, symmetrical=False, **kwargs):
		if isinstance(origin, Node):
			origin = origin.name
		if isinstance(destination, Node):
			destination = destination.name
		self.add_portal(origin, destination, symmetrical, **kwargs)
		return self.engine._get_edge(self, origin, destination, 0)

	def add_portals_from(self, seq, symmetrical=False):
		"""Make portals for a sequence of (origin, destination) pairs

        Actually, triples are acceptable too, in which case the third
        item is a dictionary of stats for the new :class:`Portal`.

        If optional argument ``symmetrical`` is set to ``True``, all
        the :class:`Portal` instances will have a mirror portal going
        in the opposite direction, which will always have the same
        stats.

        """
		for tup in seq:
			orig = tup[0]
			dest = tup[1]
			kwargs = tup[2] if len(tup) > 2 else {}
			if symmetrical:
				kwargs['symmetrical'] = True
			self.add_portal(orig, dest, **kwargs)

	def add_unit(self, a, b=None):
		"""Start keeping track of a unit"""
		if self.engine._planning:
			raise NotImplementedError(
				"Currently can't add units within a plan")
		if b is None:
			if not (isinstance(a, Place) or isinstance(a, Thing)):
				raise TypeError('when called with one argument, '
								'it must be a place or thing')
			g = a.character.name
			n = a.name
		else:
			if isinstance(a, Character):
				g = a.name
			elif not isinstance(a, str):
				raise TypeError('when called with two arguments, '
								'the first is a character or its name')
			else:
				g = a
			if isinstance(b, Place) or isinstance(b, Thing):
				n = b.name
			elif not isinstance(b, str):
				raise TypeError('when called with two arguments, '
								'the second is a thing/place or its name')
			else:
				n = b
		# This will create the node if it doesn't exist. Otherwise
		# it's redundant but harmless.
		self.engine._exist_node(g, n)
		# Declare that the node is my unit
		branch, turn, tick = self.engine._nbtt()
		self.engine._remember_unitness(self.name,
										g,
										n,
										branch=branch,
										turn=turn,
										tick=tick)

	def remove_unit(self, a, b=None):
		"""This is no longer my unit, though it still exists"""
		if self.engine._planning:
			raise NotImplementedError(
				"Currently can't remove units within a plan")
		if b is None:
			if not isinstance(a, Node):
				raise TypeError("In single argument form, "
								"del_unit requires a Node object "
								"(Thing or Place).")
			g = a.character.name
			n = a.name
		else:
			g = a.name if isinstance(a, Character) else a
			n = b.name if isinstance(b, Node) else b
		self.engine._remember_unitness(self.character.name, g, n, False)

	def portals(self):
		"""Iterate over all portals."""
		char = self.character
		make_edge = self.engine._get_edge
		for (o, d) in self.engine._edges_cache.iter_keys(
			self.character.name, *self.engine._btt()):
			yield make_edge(char, o, d)

	def units(self):
		"""Iterate over all my units

        Regardless of what character they are in.

        """
		charname = self.character.name
		branch, turn, tick = self.engine._btt()
		charmap = self.engine.character
		avit = self.engine._unitness_cache.iter_entities
		makenode = self.engine._get_node
		for graph in avit(charname, branch, turn, tick):
			for node in avit(charname, graph, branch, turn, tick):
				try:
					yield makenode(charmap[graph], node)
				except KeyError:
					continue

	def historical(self, stat):
		"""Get a historical view on the given stat

        This functions like the value of the stat, but changes
        when you time travel. Comparisons performed on the
        historical view can be passed to ``engine.turns_when``
        to find out when the comparison held true.

        """
		return StatusAlias(entity=self.stat, stat=stat, engine=self.engine)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""A base class for nodes that can be in a character.

Every actual node that you're meant to use will be a place or
thing. This module is for what they have in common.

"""
from collections.abc import Mapping, ValuesView

from networkx import shortest_path, shortest_path_length

from .allegedb import graph, HistoricKeyError

from .util import getatt
from .query import StatusAlias
from . import rule
from .exc import AmbiguousUserError


class UserMapping(Mapping):
	"""A mapping of the characters that have a particular node as an avatar.

    Getting characters from here isn't any better than getting them from
    the engine direct, but with this you can do things like use the
    .get() method to get a character if it's a user and otherwise
    get something else; or test whether the character's name is in
    the keys; and so on.

    """
	__slots__ = ['node']

	def __init__(self, node):
		"""Store the node"""
		self.node = node

	engine = getatt('node.engine')

	def _user_names(self):
		node = self.node
		engine = self.engine
		charn = node.character.name
		nn = node.name
		cache = engine._unitness_cache.user_order
		if charn not in cache or nn not in cache[charn]:
			return
		cache = cache[charn][nn]
		seen = set()
		for user in cache:
			if user in seen:
				continue
			for (branch, turn, tick) in engine._iter_parent_btt():
				if branch in cache[user]:
					branchd = cache[user][branch]
					try:
						if turn in branchd:
							if branchd[turn].get(tick, False):
								yield user
						elif branchd.rev_gettable(turn):
							turnd = branchd[turn]
							if turnd[turnd.end]:
								yield user
						seen.add(user)
						break
					except HistoricKeyError as ex:
						if ex.deleted:
							break

	@property
	def only(self):
		if len(self) != 1:
			raise AmbiguousUserError("No users, or more than one")
		return next(iter(self.values()))

	def __iter__(self):
		yield from self._user_names()

	def __len__(self):
		n = 0
		for user in self._user_names():
			n += 1
		return n

	def __bool__(self):
		for user in self._user_names():
			return True
		return False

	def __contains__(self, item):
		if item in self.engine.character:
			item = self.engine.character[item]
		if hasattr(item, 'unit'):
			charn = self.node.character.name
			nn = self.node.name
			return charn in item.unit and nn in item.unit[charn]
		return False

	def __getitem__(self, k):
		ret = self.engine.character[k]
		node = self.node
		charn = node.character.name
		nn = node.name
		avatar = ret.unit
		if charn not in avatar or nn not in avatar[charn]:
			raise KeyError("{} not used by {}".format(self.node.name, k))
		return ret


class NodeContentValues(ValuesView):
	_mapping: 'NodeContent'

	def __iter__(self):
		node = self._mapping.node
		nodem = node.character.node
		try:
			conts = node.engine._node_contents(node.character.name, node.name)
		except KeyError:
			return
		for name in conts:
			if name not in nodem:
				return
			yield nodem[name]

	def __contains__(self, item):
		return item.location == self._mapping.node


class NodeContent(Mapping):
	__slots__ = ('node', )

	def __init__(self, node):
		self.node = node

	def __iter__(self):
		try:
			it = self.node.engine._node_contents_cache.retrieve(
				self.node.character.name, self.node.name,
				*self.node.engine._btt())
		except KeyError:
			return
		yield from it

	def __len__(self):
		try:
			return len(
				self.node.engine._node_contents_cache.retrieve(
					self.node.character.name, self.node.name,
					*self.node.engine._btt()))
		except KeyError:
			return 0

	def __contains__(self, item):
		try:
			return self.node.character.thing[item].location == self.node
		except KeyError:
			return False

	def __getitem__(self, item):
		if item not in self:
			raise KeyError
		return self.node.character.thing[item]

	def values(self):
		return NodeContentValues(self)


class DestsValues(ValuesView):
	_mapping: 'Dests'

	def __contains__(self, item):
		_, name = self._mapping._pn
		return item.origin.name == name


class Dests(Mapping):
	__slots__ = ('_ecnb', '_pn')

	def __init__(self, node):
		name = node.name
		character = node.character
		engine = node.engine
		self._pn = (character.portal, name)
		self._ecnb = (engine._edges_cache, character.name, name, engine._btt)

	def __iter__(self):
		edges_cache, charname, name, btt = self._ecnb
		yield from edges_cache.iter_successors(charname, name, *btt())

	def __len__(self):
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.count_successors(charname, name, *btt())

	def __contains__(self, item):
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.has_successor(charname, name, item, *btt())

	def __getitem__(self, item):
		portal, name = self._pn
		return portal[name][item]

	def values(self):
		return DestsValues(self)


class OrigsValues(ValuesView):
	_mapping: 'Origs'

	def __contains__(self, item):
		_, name = self._mapping._pn
		return item.destination.name == name


class Origs(Mapping):
	__slots__ = ('_pn', '_ecnb')

	def __init__(self, node):
		name = node.name
		character = node.character
		engine = node.engine
		self._pn = (character.portal, name)
		self._ecnb = (engine._edges_cache, character.name, name, engine._btt)

	def __iter__(self):
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.iter_predecessors(charname, name, *btt())

	def __contains__(self, item):
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.has_predecessor(charname, name, item, *btt())

	def __len__(self):
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.count_predecessors(charname, name, *btt())

	def __getitem__(self, item):
		if item not in self:
			raise KeyError
		portal, name = self._pn
		return portal[item][name]

	def values(self):
		return OrigsValues(self)


class Node(graph.Node, rule.RuleFollower):
	"""The fundamental graph component, which edges (in LiSE, "portals")
    go between.

    Every LiSE node is either a thing or a place. They share in common
    the abilities to follow rules; to be connected by portals; and to
    contain things.

    This is truthy if it exists, falsy if it's been deleted.

    """
	__slots__ = ('_real_rule_mapping', )
	character = getatt('graph')
	name = getatt('node')
	no_unwrap = True
	_extra_keys = {
		'name',
	}

	def _get_rule_mapping(self):
		return rule.RuleMapping(self.db, self.rulebook)

	def _get_rulebook_name(self):
		try:
			return self.engine._nodes_rulebooks_cache.retrieve(
				self.character.name, self.name, *self.engine._btt())
		except KeyError:
			return self.character.name, self.name

	def _get_rulebook(self):
		return rule.RuleBook(self.engine, self._get_rulebook_name())

	def _set_rulebook_name(self, rulebook):
		character = self.character.name
		node = self.name
		cache = self.engine._nodes_rulebooks_cache
		try:
			if rulebook == cache.retrieve(character, node,
											*self.engine._btt()):
				return
		except KeyError:
			pass
		branch, turn, tick = self.engine._nbtt()
		cache.store(character, node, branch, turn, tick, rulebook)
		self.engine.query.set_node_rulebook(character, node, branch, turn,
											tick, rulebook)

	successor = adj = edge = getatt('portal')
	predecessor = pred = getatt('preportal')
	engine = getatt('db')

	@property
	def user(self):
		return UserMapping(self)

	def __init__(self, character, name, clobber=False):
		"""Store character and name, and initialize caches"""
		super().__init__(character, name)
		self.db = character.engine

	@property
	def portal(self):
		return Dests(self)

	@property
	def preportal(self):
		return Origs(self)

	def __iter__(self):
		yield from super().__iter__()
		yield from self._extra_keys
		return

	def clear(self):
		for key in super().__iter__():
			del self[key]

	def __contains__(self, k):
		"""Handle extra keys, then delegate."""
		return k in self._extra_keys or super().__contains__(k)

	def __setitem__(self, k, v):
		super().__setitem__(k, v)
		self.send(self, key=k, val=v)

	def __delitem__(self, k):
		super().__delitem__(k)
		self.send(self, key=k, val=None)

	def portals(self):
		"""Iterate over :class:`Portal` objects that lead away from me"""
		yield from self.portal.values()

	def successors(self):
		"""Iterate over nodes with edges leading from here to there."""
		for port in self.portal.values():
			yield port.destination

	def preportals(self):
		"""Iterate over :class:`Portal` objects that lead to me"""
		yield from self.preportal.values()

	def predecessors(self):
		"""Iterate over nodes with edges leading here from there."""
		for port in self.preportal.values():
			yield port.origin

	def _plain_dest_name(self, dest):
		if isinstance(dest, Node):
			if dest.character != self.character:
				raise ValueError("{} not in {}".format(dest.name,
														self.character.name))
			return dest.name
		else:
			if dest in self.character.node:
				return dest
			raise ValueError("{} not in {}".format(dest, self.character.name))

	def shortest_path_length(self, dest, weight=None):
		"""Return the length of the path from me to ``dest``.

        Raise ``ValueError`` if ``dest`` is not a node in my character
        or the name of one.

        """

		return shortest_path_length(self.character, self.name,
									self._plain_dest_name(dest), weight)

	def shortest_path(self, dest, weight=None):
		"""Return a list of node names leading from me to ``dest``.

        Raise ``ValueError`` if ``dest`` is not a node in my character
        or the name of one.

        """
		return shortest_path(self.character, self.name,
								self._plain_dest_name(dest), weight)

	def path_exists(self, dest, weight=None):
		"""Return whether there is a path leading from me to ``dest``.

        With ``weight``, only consider edges that have a stat by the
        given name.

        Raise ``ValueError`` if ``dest`` is not a node in my character
        or the name of one.

        """
		try:
			return bool(self.shortest_path_length(dest, weight))
		except KeyError:
			return False

	@property
	def content(self):
		return NodeContent(self)

	def contents(self):
		return self.content.values()

	def delete(self):
		"""Get rid of this, starting now.

        Apart from deleting the node, this also informs all its users
        that it doesn't exist and therefore can't be their avatar
        anymore.

        """
		self.clear()
		for contained in list(self.contents()):
			contained.delete()
		if self.name in self.character.portal:
			del self.character.portal[self.name]
		if self.name in self.character.preportal:
			del self.character.preportal[self.name]
		for user in list(self.user.values()):
			user.remove_unit(self.character.name, self.name)
		branch, turn, tick = self.engine._nbtt()
		self.engine._nodes_cache.store(self.character.name, self.name, branch,
										turn, tick, False)
		self.engine.query.exist_node(self.character.name, self.name, branch,
										turn, tick, False)
		self.character.node.send(self.character.node, key=self.name, val=None)

	def one_way_portal(self, other, **stats):
		"""Connect a portal from here to another node, and return it."""
		return self.character.new_portal(self,
											other,
											symmetrical=False,
											**stats)

	def one_way(self, other, **stats):
		"""Connect a portal from here to another node, and return it."""
		return self.one_way_portal(other, **stats)

	def two_way_portal(self, other, **stats):
		"""Connect these nodes with a two-way portal and return it."""
		return self.character.new_portal(self,
											other,
											symmetrical=True,
											**stats)

	def two_way(self, other, **stats):
		"""Connect these nodes with a two-way portal and return it."""
		return self.two_way_portal(other, **stats)

	def new_thing(self, name, **stats):
		"""Create a new thing, located here, and return it."""
		return self.character.new_thing(name, self.name, **stats)

	def historical(self, stat):
		"""Return a reference to the values that a stat has had in the past.

        You can use the reference in comparisons to make a history
        query, and execute the query by calling it, or passing it to
        ``self.engine.ticks_when``.

        """
		return StatusAlias(entity=self, stat=stat)

	def __bool__(self):
		return self.name in self.character.node



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""A dwarf hunting a kobold that hides in the bushes.

This script will initialize world.db and the code libraries to run the
simulation described. To view it, run ELiDE from the same directory
as you ran this script from.

"""

import networkx as nx


def inittest(engine,
				mapsize=(10, 10),
				dwarf_pos=(0, 0),
				kobold_pos=(9, 9),
				shrubberies=20,
				dwarf_sight_radius=2,
				kobold_sprint_chance=.1):
	# initialize world
	phys = engine.new_character('physical', data=nx.grid_2d_graph(*mapsize))
	assert len(phys.place) == mapsize[0] * mapsize[1]
	kobold = phys.new_thing("kobold", kobold_pos)
	assert len(phys.thing) == 1
	kobold['sprint_chance'] = kobold_sprint_chance
	kobold['_image_paths'] = ['atlas://rltiles/base.atlas/kobold_m']
	dwarf = phys.new_thing("dwarf", dwarf_pos)
	dwarf['sight_radius'] = dwarf_sight_radius
	dwarf['seen_kobold'] = False
	dwarf['_image_paths'] = ['atlas://rltiles/base.atlas/dwarf_m']
	# randomly place the shrubberies and add their locations to shrub_places
	locs = list(phys.place.keys())
	engine.shuffle(locs)
	shrub_places = []
	while len(shrub_places) < shrubberies:
		loc = locs.pop()
		phys.add_thing("shrub" + str(len(shrub_places)),
						loc,
						cover=1,
						_image_paths=['atlas://rltiles/dc-mon.atlas/fungus'],
						_group='shrub')
		shrub_places.append(loc)
	print('{} shrubberies: {}'.format(len(shrub_places), shrub_places))
	kobold['shrub_places'] = shrub_places

	# Basic day night cycle
	@engine.action
	def time_slipping(character,
						*,
						daylen: int,
						nightlen: int,
						twilightlen: float = 0.0):
		if 'hour' not in character.stat:
			character.stat['hour'] = 0
			character.stat['day_period'] = 'dawn' if twilightlen else 'day'
			return
		twi_margin = twilightlen / 2
		hour = character.stat['hour'] = (character.stat['hour'] +
											1) % (daylen + nightlen)
		if twilightlen:
			if hour < twi_margin or hour > daylen + nightlen - twi_margin:
				character.stat['day_period'] = 'dawn'
			elif twi_margin < hour < daylen - twi_margin:
				character.stat['day_period'] = 'day'
			elif daylen - twi_margin < hour < daylen + twi_margin:
				character.stat['day_period'] = 'dusk'
			else:
				character.stat['day_period'] = 'night'
		else:
			character.stat['day_period'] = 'day' if hour < daylen else 'night'

	# If the kobold is not in a shrubbery, it will try to get to one.
	# If it is, there's a chance it will try to get to another.
	@kobold.rule
	def shrubsprint(thing):
		shrub_places = sorted(list(thing['shrub_places']))
		if thing['location'] in shrub_places:
			shrub_places.remove(thing['location'])
			assert thing['location'] not in shrub_places
		whereto = thing.engine.choice(shrub_places)
		thing.travel_to(whereto)

	@shrubsprint.trigger
	def uncovered(thing):
		for shrub_candidate in thing.location.contents():
			if shrub_candidate.name[:5] == "shrub":
				return False
		thing.engine.info("kobold uncovered")
		return True

	@shrubsprint.trigger
	def breakcover(thing):
		if thing.engine.random() < thing['sprint_chance']:
			thing.engine.info("kobold breaking cover")
			return True

	@shrubsprint.prereq
	def not_traveling(thing):
		if thing.next_location is not None:
			thing.engine.info("kobold already travelling to {}".format(
				thing.next_location))
			return False
		else:
			return True

	@engine.method
	def set_kill_flag(eng):
		eng.character['physical'].thing['dwarf']['kill'] = True

	@dwarf.rule
	def fight(thing):
		method = thing.engine.method
		return "Kill kobold?", [("Kill", method.set_kill_flag),
								("Spare", None)]

	@fight.trigger
	def sametile(thing):
		try:
			return (thing['location'] == thing.character.thing['kobold']
					['location'])
		except KeyError:
			return False

	@fight.prereq
	def kobold_alive(thing):
		return 'kobold' in thing.character.thing

	def aware(thing):
		# calculate the distance from dwarf to kobold
		from math import hypot
		try:
			bold = thing.character.thing['kobold']
		except KeyError:
			return False
		(dx, dy) = bold['location']
		(ox, oy) = thing['location']
		xdist = abs(dx - ox)
		ydist = abs(dy - oy)
		dist = hypot(xdist, ydist)
		# if it's <= the dwarf's sight radius, the dwarf is aware of the kobold
		return dist <= thing['sight_radius']

	fight.prereq(aware)

	@dwarf.rule
	def kill_kobold(thing):
		del thing.character.thing['kobold']
		del thing['kill']

	kill_kobold.trigger(kobold_alive)

	@kill_kobold.prereq
	def unmerciful(thing):
		return thing.get('kill', False)

	@dwarf.rule
	def go2kobold(thing):
		thing.travel_to(thing.character.thing['kobold']['location'])

	go2kobold.trigger(aware)

	go2kobold.prereqs = ['kobold_alive']

	@go2kobold.prereq
	def kobold_not_here(thing):
		return 'kobold' not in thing.location.content

	@dwarf.rule
	def wander(thing):
		dests = sorted(list(thing.character.place.keys()))
		dests.remove(thing['location'])
		thing.travel_to(thing.engine.choice(dests))

	@wander.trigger
	def standing_still(thing):
		return thing.next_location is None


if __name__ == '__main__':
	from LiSE.engine import Engine
	from os import remove

	with Engine(random_seed=69105, clear=True) as engine:
		inittest(engine, shrubberies=20, kobold_sprint_chance=.9)
		engine.commit()
		print('shrub_places beginning: {}'.format(
			engine.character['physical'].thing['kobold']['shrub_places']))



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Directed edges, as used by LiSE."""
from collections.abc import Mapping
from typing import Union, List, Tuple, Any

from .allegedb.graph import Edge
from .allegedb import HistoricKeyError

from .util import getatt
from .query import StatusAlias
from .rule import RuleFollower
from .rule import RuleMapping as BaseRuleMapping


class RuleMapping(BaseRuleMapping):
	"""Mapping to get rules followed by a portal."""

	def __init__(self, portal):
		"""Store portal, engine, and rulebook."""
		super().__init__(portal.engine, portal.rulebook)
		self.portal = portal


class Portal(Edge, RuleFollower):
	"""Connection between two Places that Things may travel along.

    Portals are one-way, but you can make one appear two-way by
    setting the ``symmetrical`` key to ``True``,
    eg. ``character.add_portal(orig, dest, symmetrical=True)``.
    The portal going the other way will appear to have all the
    stats of this one, and attempting to set a stat on it will
    set it here instead.

    LiSE entities are truthy so long as they exist, falsy if they've
    been deleted.

    """
	__slots__ = ('graph', 'orig', 'dest', 'idx', 'origin', 'destination',
					'_rulebook', '_real_rule_mapping')
	character = getatt('graph')
	engine = getatt('db')
	no_unwrap = True

	def __init__(self, graph, orig, dest, idx=0):
		super().__init__(graph, orig, dest, idx)
		self.origin = graph.node[orig]
		self.destination = graph.node[dest]

	@property
	def _cache(self):
		return self.db._edge_val_cache[self.character.name][self.orig][
			self.dest][0]

	def _rule_name_activeness(self):
		rulebook_name = self._get_rulebook_name()
		cache = self.engine._active_rules_cache
		if rulebook_name not in cache:
			return
		cache = cache[rulebook_name]
		for rule in cache:
			for (branch, turn, tick) in self.engine._iter_parent_btt():
				if branch not in cache[rule]:
					continue
				try:
					yield (rule, cache[rule][branch][turn][tick])
					break
				except ValueError:
					continue
				except HistoricKeyError as ex:
					if ex.deleted:
						break
		raise KeyError("{}->{} has no rulebook?".format(self.orig, self.dest))

	def _get_rulebook_name(self):
		try:
			return self.engine._portals_rulebooks_cache.retrieve(
				self.character.name, self.orig, self.dest, *self.engine._btt())
		except KeyError:
			return (self.character.name, self.orig, self.dest)

	def _set_rulebook_name(self, rulebook):
		character = self.character
		orig = self.orig
		dest = self.dest
		cache = self.engine._portals_rulebooks_cache
		try:
			if rulebook == cache.retrieve(character, orig, dest,
											*self.engine._btt()):
				return
		except KeyError:
			pass
		branch, turn, tick = self.engine._nbtt()
		cache.store(character, orig, dest, branch, turn, tick, rulebook)
		self.engine.query.set_portal_rulebook(character, orig, dest, branch,
												turn, tick, rulebook)

	def _get_rule_mapping(self):
		return RuleMapping(self)

	def __getitem__(self, key):
		"""Get the present value of the key.

        If I am a mirror of another Portal, return the value from that
        Portal instead.

        """
		if key == 'origin':
			return self.orig
		elif key == 'destination':
			return self.dest
		elif key == 'character':
			return self.character.name
		elif key == 'is_mirror':
			try:
				return super().__getitem__(key)
			except KeyError:
				return False
		elif 'is_mirror' in self and self['is_mirror']:
			return self.character.preportal[self.orig][self.dest][key]
		else:
			return super().__getitem__(key)

	def __setitem__(self, key, value):
		"""Set ``key``=``value`` at the present game-time.

        If I am a mirror of another Portal, set ``key``=``value`` on
        that Portal instead.

        """
		if key in ('origin', 'destination', 'character'):
			raise KeyError("Can't change " + key)
		elif 'is_mirror' in self and self['is_mirror']:
			self.reciprocal[key] = value
			return
		elif key == 'symmetrical' and value:
			if (self.dest not in self.character.portal
				or self.orig not in self.character.portal[self.dest]):
				self.character.add_portal(self.dest, self.orig)
			self.character.portal[self.dest][self.orig]["is_mirror"] = True
			self.send(self, key='symmetrical', val=False)
			return
		elif key == 'symmetrical' and not value:
			try:
				self.character.portal[self.dest][
					self.orig]["is_mirror"] = False
			except KeyError:
				pass
			self.send(self, key='symmetrical', val=False)
			return
		super().__setitem__(key, value)

	def __repr__(self):
		"""Describe character, origin, and destination"""
		return "<{}.character[{}].portal[{}][{}]>".format(
			repr(self.engine), repr(self['character']), repr(self['origin']),
			repr(self['destination']))

	def __bool__(self):
		"""It means something that I exist, even if I have no data."""
		return self.orig in self.character.portal and self.dest in self.character.portal[
			self.orig]

	@property
	def reciprocal(self):
		"""If there's another Portal connecting the same origin and
        destination that I do, but going the opposite way, return
        it. Else raise KeyError.

        """
		try:
			return self.character.portal[self.dest][self.orig]
		except KeyError:
			raise KeyError("This portal has no reciprocal")

	def historical(self, stat):
		"""Return a reference to the values that a stat has had in the past.

        You can use the reference in comparisons to make a history
        query, and execute the query by calling it, or passing it to
        ``self.engine.ticks_when``.

        """
		return StatusAlias(entity=self, stat=stat)

	def update(self, e: Union[Mapping, List[Tuple[Any, Any]]] = None, **f):
		"""Works like regular update, but only actually updates when the new
        value and the old value differ. This is necessary to prevent
        certain infinite loops.

        """
		if e is not None:
			if hasattr(e, 'keys') and callable(e.keys):
				for k in e.keys():
					if k not in self:
						self[k] = e[k]
					else:
						v = e[k]
						if self[k] != v:
							self[k] = v
			else:
				for k, v in e:
					if k not in self or self[k] != v:
						self[k] = v
		for k, v in f.items():
			if k not in self or self[k] != v:
				self[k] = v

	def delete(self):
		"""Remove myself from my :class:`Character`.

        For symmetry with :class:`Thing` and :class:`Place`.

        """
		self.clear()
		branch, turn, tick = self.engine._nbtt()
		self.engine._edges_cache.store(self.character.name, self.origin.name,
										self.destination.name, 0, branch, turn,
										tick, None)
		self.engine.query.exist_edge(self.character.name, self.origin.name,
										self.destination.name, branch, turn,
										tick, False)
		try:
			del self.engine._edge_objs[(self.graph.name, self.orig, self.dest)]
		except KeyError:
			pass
		self.character.portal[self.origin.name].send(
			self.character.portal[self.origin.name], key='dest', val=None)

	def unwrap(self):
		return {
			k: v.unwrap()
			if hasattr(v, 'unwrap') and not hasattr(v, 'no_unwrap') else v
			for (k, v) in self.items()
		}



# This file is part of allegedb, a database abstraction for versioned graphs
# Copyright (c) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Classes for in-memory storage and retrieval of historical graph data.
"""
from .window import WindowDict, HistoricKeyError, FuturistWindowDict, \
 TurnDict, SettingsTurnDict
from collections import OrderedDict, defaultdict, deque


def _default_args_munger(self, k):
	"""By default, `PickyDefaultDict`'s ``type`` takes no positional arguments.

    """
	return tuple()


def _default_kwargs_munger(self, k):
	"""By default, `PickyDefaultDict`'s ``type`` takes no keyword arguments."""
	return {}


class PickyDefaultDict(dict):
	"""A ``defaultdict`` alternative that requires values of a specific type.

    Pass some type object (such as a class) to the constructor to
    specify what type to use by default, which is the only type I will
    accept.

    Default values are constructed with no arguments by default;
    supply ``args_munger`` and/or ``kwargs_munger`` to override this.
    They take arguments ``self`` and the unused key being looked up.

    """
	__slots__ = ['type', 'args_munger', 'kwargs_munger', 'parent', 'key']

	def __init__(self,
					type,
					args_munger=_default_args_munger,
					kwargs_munger=_default_kwargs_munger):
		self.type = type
		self.args_munger = args_munger
		self.kwargs_munger = kwargs_munger

	def __getitem__(self, k):
		if k in self:
			return super(PickyDefaultDict, self).__getitem__(k)
		try:
			ret = self[k] = self.type(*self.args_munger(self, k),
										**self.kwargs_munger(self, k))
		except TypeError:
			raise KeyError(k)
		return ret

	def _create(self, v):
		return self.type(v)

	def __setitem__(self, k, v):
		if type(v) is not self.type:
			v = self._create(v)
		super(PickyDefaultDict, self).__setitem__(k, v)


class StructuredDefaultDict(dict):
	"""A `defaultdict`-like class with values stored at a specific depth.

    Requires an integer to tell it how many layers deep to go.
    The innermost layer will be ``PickyDefaultDict``, which will take the
    ``type``, ``args_munger``, and ``kwargs_munger`` arguments supplied
    to my constructor.

    """
	__slots__ = ('layer', 'type', 'args_munger', 'kwargs_munger', 'parent',
					'key', '_stuff', 'gettest', 'settest')

	def __init__(self,
					layers,
					type=None,
					args_munger=_default_args_munger,
					kwargs_munger=_default_kwargs_munger,
					gettest=lambda k: None,
					settest=lambda k, v: None):
		if layers < 1:
			raise ValueError("Not enough layers")
		self.layer = layers
		self.type = type
		self.args_munger = args_munger
		self.kwargs_munger = kwargs_munger
		self._stuff = (layers, type, args_munger, kwargs_munger)
		self.gettest = gettest
		self.settest = settest

	def __getitem__(self, k):
		self.gettest(k)
		if k in self:
			return dict.__getitem__(self, k)
		layer, typ, args_munger, kwargs_munger = self._stuff
		if layer == 1:
			if typ is None:
				ret = {}
			else:
				ret = PickyDefaultDict(typ, args_munger, kwargs_munger)
				ret.parent = self
				ret.key = k
		elif layer < 1:
			raise ValueError("Invalid layer")
		else:
			ret = StructuredDefaultDict(layer - 1, typ, args_munger,
										kwargs_munger)
			ret.parent = self
			ret.key = k
		dict.__setitem__(self, k, ret)
		return ret

	def __setitem__(self, k, v):
		self.settest(k, v)
		if type(v) is StructuredDefaultDict:
			layer, typ, args_munger, kwargs_munger = self._stuff
			if (v.layer == layer - 1 and (typ is None or v.type is typ)
				and v.args_munger is args_munger
				and v.kwargs_munger is kwargs_munger):
				super().__setitem__(k, v)
				return
		elif type(v) is PickyDefaultDict:
			layer, typ, args_munger, kwargs_munger = self._stuff
			if (layer == 1 and v.type is typ and v.args_munger is args_munger
				and v.kwargs_munger is kwargs_munger):
				super().__setitem__(k, v)
				return
		raise TypeError("Can't set layer {}".format(self.layer))


KEYCACHE_MAXSIZE = 1024


def lru_append(kc, lru, kckey, maxsize):
	"""Delete old data from ``kc``, then add new ``kckey`` to ``lru``

    :param kc: a three-layer keycache
    :param lru: an :class:`OrderedDict` with a key for each triple that should
                fill out ``kc``'s three layers
    :param kckey: a triple that indexes into ``kc``, which will be added to
                  ``lru`` if needed
    :param maxsize: maximum number of entries in ``lru`` and, therefore, ``kc``

    """
	if kckey in lru:
		return
	while len(lru) >= maxsize:
		(peb, turn, tick), _ = lru.popitem(False)
		if peb not in kc:
			continue
		kcpeb = kc[peb]
		if turn not in kcpeb:
			continue
		kcpebturn = kcpeb[turn]
		if tick not in kcpebturn:
			continue
		del kcpebturn[tick]
		if not kcpebturn:
			del kcpeb[turn]
		if not kcpeb:
			del kc[peb]
	lru[kckey] = True


class Cache:
	"""A data store that's useful for tracking graph revisions."""
	__slots__ = ('db', 'parents', 'keys', 'keycache', 'branches', 'shallowest',
					'settings', 'presettings', 'time_entity', '_kc_lru',
					'_store_stuff', '_remove_stuff', '_truncate_stuff',
					'setdb', 'deldb', 'keyframe', 'name')

	def __init__(self, db, kfkvs=None):
		super().__init__()
		self.db = db
		self.parents = StructuredDefaultDict(3, SettingsTurnDict)
		"""Entity data keyed by the entities' parents.

        An entity's parent is what it's contained in. When speaking of a node,
        this is its graph. When speaking of an edge, the parent is usually the
        graph and the origin in a pair, though for multigraphs the destination
        might be part of the parent as well.

        Deeper layers of this cache are keyed by branch and revision.

        """
		self.keys = StructuredDefaultDict(2, SettingsTurnDict)
		"""Cache of entity data keyed by the entities themselves.

        That means the whole tuple identifying the entity is the
        top-level key in this cache here. The second-to-top level
        is the key within the entity.

        Deeper layers of this cache are keyed by branch, turn, and tick.

        """
		self.keycache = PickyDefaultDict(SettingsTurnDict)
		"""Keys an entity has at a given turn and tick."""
		self.branches = StructuredDefaultDict(1, SettingsTurnDict)
		"""A less structured alternative to ``keys``.

        For when you already know the entity and the key within it,
        but still need to iterate through history to find the value.

        """
		self.keyframe = StructuredDefaultDict(1, SettingsTurnDict,
												**(kfkvs or {}))
		"""Key-value dictionaries representing my state at a given time"""
		self.shallowest = OrderedDict()
		"""A dictionary for plain, unstructured hinting."""
		self.settings = PickyDefaultDict(SettingsTurnDict)
		"""All the ``entity[key] = value`` settings on some turn"""
		self.presettings = PickyDefaultDict(SettingsTurnDict)
		"""The values prior to ``entity[key] = value`` settings on some turn"""
		self.time_entity = {}
		self._kc_lru = OrderedDict()
		self._store_stuff = (self.parents, self.branches, self.keys,
								db.delete_plan, db._time_plan,
								self._iter_future_contradictions, db._branches,
								db._turn_end, self._store_journal,
								self.time_entity, db._where_cached,
								self.keycache, db, self._update_keycache)
		self._remove_stuff = (self.time_entity, self.parents, self.branches,
								self.keys, self.settings, self.presettings,
								self._remove_keycache, self.keycache)
		self._truncate_stuff = (self.parents, self.branches, self.keys,
								self.settings, self.presettings, self.keycache)

	def load(self, data):
		"""Add a bunch of data. Must be in chronological order.

        But it doesn't need to all be from the same branch, as long as
        each branch is chronological of itself.

        """
		branches = defaultdict(list)
		for row in data:
			branches[row[-4]].append(row)
		db = self.db
		# Make keycaches and valcaches. Must be done chronologically
		# to make forwarding work.
		childbranch = db._childbranch
		branch2do = deque(['trunk'])

		store = self.store
		while branch2do:
			branch = branch2do.popleft()
			for row in branches[branch]:
				store(*row, planning=False, loading=True)
			if branch in childbranch:
				branch2do.extend(childbranch[branch])

	def _valcache_lookup(self, cache, branch, turn, tick):
		"""Return the value at the given time in ``cache``"""
		if branch in cache:
			branc = cache[branch]
			try:
				if turn in branc and branc[turn].rev_gettable(tick):
					return branc[turn][tick]
				elif branc.rev_gettable(turn - 1):
					turnd = branc[turn - 1]
					return turnd.final()
			except HistoricKeyError as ex:
				# probably shouldn't ever happen, empty branches
				# shouldn't be kept in the cache at all...
				# but it's easy to handle
				if ex.deleted:
					raise
		for b, r, t in self.db._iter_parent_btt(branch, turn, tick):
			if b in cache:
				if r in cache[b] and cache[b][r].rev_gettable(t):
					try:
						return cache[b][r][t]
					except HistoricKeyError as ex:
						if ex.deleted:
							raise
				elif cache[b].rev_gettable(r - 1):
					cbr = cache[b][r - 1]
					try:
						return cbr.final()
					except HistoricKeyError as ex:
						if ex.deleted:
							raise

	def _get_keycachelike(self, keycache, keys, get_adds_dels, parentity,
							branch, turn, tick, *, forward):
		"""Try to retrieve a frozenset representing extant keys.

        If I can't, generate one, store it, and return it.

        """
		keycache_key = parentity + (branch, )
		keycache2 = keycache3 = None
		if keycache_key in keycache:
			keycache2 = keycache[keycache_key]
			if turn in keycache2:
				keycache3 = keycache2[turn]
				if tick in keycache3:
					return keycache3[tick]
		if forward:
			# Take valid values from the past of a keycache and copy them
			# forward, into the present. Assumes that time is only moving
			# forward, never backward, never skipping any turns or ticks,
			# and any changes to the world state are happening through
			# allegedb proper, meaning they'll all get cached. In LiSE this
			# means every change to the world state should happen inside of
			# a call to ``Engine.next_turn`` in a rule.
			if keycache2 and keycache2.rev_gettable(turn):
				# there's a keycache from a prior turn in this branch. Get it
				if turn not in keycache2:
					# since it's not this *exact* turn, there might be changes
					old_turn = keycache2.rev_before(turn)
					old_turn_kc = keycache2[turn]
					added, deleted = get_adds_dels(
						parentity,
						branch,
						turn,
						tick,
						stoptime=(branch, old_turn, old_turn_kc.end),
						cache=keys)
					ret = old_turn_kc.final().union(added).difference(deleted)
					# assert ret == get_adds_dels(
					# keys[parentity], branch, turn, tick)[0]  # slow
					new_turn_kc = WindowDict()
					new_turn_kc[tick] = ret
					keycache2[turn] = new_turn_kc
					return ret
				if not keycache3:
					keycache3 = keycache2[turn]
				if tick not in keycache3:
					if keycache3.rev_gettable(tick):
						added, deleted = get_adds_dels(
							parentity,
							branch,
							turn,
							tick,
							stoptime=(branch, turn,
										keycache3.rev_before(tick)),
							cache=keys)
						ret = keycache3[tick].union(added).difference(deleted)
						# assert ret == get_adds_dels(
						# keys[parentity], branch, turn, tick)[0]  # slow
						keycache3[tick] = ret
						return ret
					else:
						turn_before = keycache2.rev_before(turn)
						tick_before = keycache2[turn_before].end
						keys_before = keycache2[turn_before][tick_before]
						added, deleted = get_adds_dels(
							parentity,
							branch,
							turn,
							tick,
							stoptime=(branch, turn_before, tick_before),
							cache=keys)
						ret = keycache3[tick] = keys_before.union(
							added).difference(deleted)
						# assert ret == get_adds_dels(
						# keys[parentity], branch, turn, tick)[0]  # slow
						return ret
				# assert kcturn[tick] == get_adds_dels(
				# keys[parentity], branch, turn, tick)[0]  # slow
				return keycache3[tick]
		ret = frozenset(get_adds_dels(parentity, branch, turn, tick)[0])
		if keycache2:
			if keycache3:
				keycache3[tick] = ret
			else:
				keycache2[turn] = {tick: ret}
		else:
			kcc = SettingsTurnDict()
			kcc[turn] = {tick: ret}
			keycache[keycache_key] = kcc
		return ret

	def _get_keycache(self, parentity, branch, turn, tick, *, forward):
		"""Get a frozenset of keys that exist in the entity at the moment.

        With ``forward=True``, enable an optimization that copies old key sets
        forward and updates them.

        """
		lru_append(self.keycache, self._kc_lru,
					(parentity + (branch, ), turn, tick), KEYCACHE_MAXSIZE)
		return self._get_keycachelike(self.keycache,
										self.keys,
										self._get_adds_dels,
										parentity,
										branch,
										turn,
										tick,
										forward=forward)

	def _update_keycache(self, *args, forward):
		"""Add or remove a key in the set describing the keys that exist."""
		entity, key, branch, turn, tick, value = args[-6:]
		parent = args[:-6]
		kc = self._get_keycache(parent + (entity, ),
								branch,
								turn,
								tick,
								forward=forward)
		if value is None:
			kc = kc.difference((key, ))
		else:
			kc = kc.union((key, ))
		self.keycache[parent + (entity, branch)][turn][tick] = kc

	def _get_adds_dels(self,
						entity,
						branch,
						turn,
						tick,
						*,
						stoptime=None,
						cache=None):
		"""Return a pair of sets describing changes to the entity's keys

        Returns a pair of sets: ``(added, deleted)``. These are the changes
        to the key set that occurred since ``stoptime``, which, if present,
        should be a triple ``(branch, turn, tick)``.

        With ``stoptime=None`` (the default), ``added`` will in fact be all
        keys, and ``deleted`` will be empty.

        """
		# Not using the journal because that doesn't distinguish entities.
		# I think I might not want to use ``stoptime`` at all, now that
		# there is such a thing as keyframes...
		cache = cache or self.keys[entity]
		added = set()
		deleted = set()
		kf = self.keyframe.get(entity, None)
		for key, branches in cache.items():
			for (branc, trn,
					tck) in self.db._iter_parent_btt(branch,
														turn,
														tick,
														stoptime=stoptime):
				if branc not in branches or not branches[branc].rev_gettable(
					trn):
					continue
				turnd = branches[branc]
				if trn in turnd:
					if turnd[trn].rev_gettable(tck):
						if turnd[trn][tck] is None:
							deleted.add(key)
						else:
							added.add(key)
						break
					else:
						trn -= 1
				if not turnd.rev_gettable(trn):
					break
				tickd = turnd[trn]
				if tickd.final() is None:
					deleted.add(key)
				else:
					added.add(key)
				break
		else:
			if stoptime or not kf:
				return added, deleted
			for (branc, trn,
					tck) in self.db._iter_parent_btt(branch, turn, tick):
				if branc not in kf or not kf[branc].rev_gettable(trn):
					continue
				kfb = kf[branc]
				if trn in kfb:
					kfbr = kfb[trn]
					if kfbr.rev_gettable(tck):
						added.update(set(kfbr[tck]).difference(deleted))
				elif kfb.rev_gettable(trn):
					added.update(set(kfb[trn].final()).difference(deleted))
		return added, deleted

	def store(self,
				*args,
				planning=None,
				forward=None,
				loading=False,
				contra=True):
		"""Put a value in various dictionaries for later .retrieve(...).

        Needs at least five arguments, of which the -1th is the value
        to store, the -2th is the tick to store it at, the -3th
        is the turn to store it in, the -4th is the branch the
        revision is in, the -5th is the key the value is for,
        and the remaining arguments identify the entity that has
        the key, eg. a graph, node, or edge.

        With ``planning=True``, you will be permitted to alter
        "history" that takes place after the last non-planning
        moment of time, without much regard to consistency.
        Otherwise, contradictions will be handled by deleting
        everything in the contradicted plan after the present moment,
        unless you set ``contra=False``.

        ``loading=True`` prevents me from updating the ORM's records
        of the ends of branches and turns.

        """
		(self_parents, self_branches, self_keys, delete_plan, time_plan,
			self_iter_future_contradictions, db_branches, db_turn_end,
			self_store_journal, self_time_entity, db_where_cached, keycache,
			db, update_keycache) = self._store_stuff
		if planning is None:
			planning = db._planning
		if forward is None:
			forward = db._forward
		entity, key, branch, turn, tick, value = args[-6:]
		parent = args[:-6]
		entikey = (entity, key)
		parentikey = parent + (entity, key)
		if parent:
			parentity = self_parents[parent][entity]
			if key in parentity:
				branches = parentity[key]
				turns = branches[branch]
			else:
				branches = self_branches[parentikey] = self_keys[
					parent + (entity, )][key] = parentity[key]
				turns = branches[branch]
		else:
			if entikey in self_branches:
				branches = self_branches[entikey]
				turns = branches[branch]
			else:
				branches = self_branches[entikey]
				self_keys[entity, ][key] = branches
				turns = branches[branch]
		if planning:
			if turn in turns and tick < turns[turn].end:
				raise HistoricKeyError(
					"Already have some ticks after {} in turn {} of branch {}".
					format(tick, turn, branch))
		if contra:
			contras = list(
				self_iter_future_contradictions(entity, key, turns, branch,
												turn, tick, value))
			if contras:
				self.shallowest = OrderedDict()
			for contra_turn, contra_tick in contras:
				if (branch, contra_turn, contra_tick
					) in time_plan:  # could've been deleted in this very loop
					delete_plan(time_plan[branch, contra_turn, contra_tick])
			if not turns:  # turns may be mutated in delete_plan
				branches[branch] = turns
			if parentikey not in self_branches:
				self_branches[parentikey] = branches
		if not loading and not planning:
			parbranch, turn_start, tick_start, turn_end, tick_end = db_branches[
				branch]
			db_branches[branch] = parbranch, turn_start, tick_start, turn, tick
			db_turn_end[branch, turn] = tick
		self_store_journal(*args)
		self.shallowest[parent + (entity, key, branch, turn, tick)] = value
		if turn in turns:
			the_turn = turns[turn]
			the_turn.truncate(tick)
			the_turn[tick] = value
		else:
			new = FuturistWindowDict()
			new[tick] = value
			turns[turn] = new
		self_time_entity[branch, turn, tick] = parent, entity, key
		where_cached = db_where_cached[args[-4:-1]]
		if self not in where_cached:
			where_cached.append(self)
		# if we're editing the past, have to invalidate the keycache
		keycache_key = parent + (entity, branch)
		if keycache_key in keycache:
			thiskeycache = keycache[keycache_key]
			if turn in thiskeycache:
				del thiskeycache[turn]
			thiskeycache.truncate(turn)
			if not thiskeycache:
				del keycache[keycache_key]
		if not db._no_kc:
			update_keycache(*args, forward=forward)

	def remove_branch(self, branch):
		time_entity, parents, branches, keys, settings, presettings, remove_keycache, keycache = self._remove_stuff
		parentikeys = set()
		for (branc, turn, tick), parentikey in list(time_entity.items()):
			if branc != branch:
				continue
			parentikeys.add(parentikey)
			del time_entity[branc, turn, tick]
		for (parent, entity, key) in parentikeys:
			branchkey = parent + (entity, key)
			keykey = parent + (entity, )
			if parent in parents:
				parentt = parents[parent]
				if entity in parentt:
					entty = parentt[entity]
					if key in entty:
						kee = entty[key]
						if branch in kee:
							del kee[branch]
						if not kee:
							del entty[key]
					if not entty:
						del parentt[entity]
				if not parentt:
					del parents[parent]
			if branchkey in branches:
				entty = branches[branchkey]
				if branch in entty:
					del entty[branch]
				if not entty:
					del branches[branchkey]
			if keykey in keys:
				entty = keys[keykey]
				if key in entty:
					kee = entty[key]
					if branch in kee:
						del kee[branch]
					if not kee:
						del entty[key]
				if not entty:
					del keys[keykey]
		del settings[branch]
		del presettings[branch]
		self.shallowest = OrderedDict()

	def remove(self, branch, turn, tick):
		"""Delete all data from a specific tick"""
		time_entity, parents, branches, keys, settings, presettings, remove_keycache, keycache = self._remove_stuff
		parent, entity, key = time_entity[branch, turn, tick]
		branchkey = parent + (entity, key)
		keykey = parent + (entity, )
		if parent in parents:
			parentt = parents[parent]
			if entity in parentt:
				entty = parentt[entity]
				if key in entty:
					kee = entty[key]
					if branch in kee:
						branhc = kee[branch]
						if turn in branhc:
							trn = branhc[turn]
							del trn[tick]
							if not trn:
								del branhc[turn]
							if not branhc:
								del kee[branch]
					if not kee:
						del entty[key]
				if not entty:
					del parentt[entity]
			if not parentt:
				del parents[parent]
		if branchkey in branches:
			entty = branches[branchkey]
			if branch in entty:
				branhc = entty[branch]
				if turn in branhc:
					trn = branhc[turn]
					if tick in trn:
						del trn[tick]
					if not trn:
						del branhc[turn]
				if not branhc:
					del entty[branch]
			if not entty:
				del branches[branchkey]
		if keykey in keys:
			entty = keys[keykey]
			if key in entty:
				kee = entty[key]
				if branch in kee:
					branhc = kee[branch]
					if turn in branhc:
						trn = entty[turn]
						if tick in trn:
							del trn[tick]
						if not trn:
							del branhc[turn]
					if not branhc:
						del kee[branch]
				if not kee:
					del entty[key]
			if not entty:
				del keys[keykey]
		branhc = settings[branch]
		pbranhc = presettings[branch]
		trn = branhc[turn]
		ptrn = pbranhc[turn]
		if tick in trn:
			del trn[tick]
		if tick in ptrn:
			del ptrn[tick]
		if not ptrn:
			del pbranhc[turn]
			del branhc[turn]
		if not pbranhc:
			del settings[branch]
			del presettings[branch]
		self.shallowest = OrderedDict()
		remove_keycache(parent + (entity, branch), turn, tick)

	def _remove_keycache(self, entity_branch, turn, tick):
		"""Remove the future of a given entity from a branch in the keycache"""
		keycache = self.keycache
		if entity_branch in keycache:
			kc = keycache[entity_branch]
			if turn in kc:
				kcturn = kc[turn]
				if tick in kcturn:
					del kcturn[tick]
				kcturn.truncate(tick)
				if not kcturn:
					del kc[turn]
			kc.truncate(turn)
			if not kc:
				del keycache[entity_branch]

	def truncate(self, branch, turn, tick, direction='forward'):
		if direction not in {'forward', 'backward'}:
			raise ValueError("Illegal direction")
		parents, branches, keys, settings, presettings, keycache = self._truncate_stuff

		def truncate_branhc(branhc):
			if turn in branhc:
				trn = branhc[turn]
				trn.truncate(tick, direction)
				branhc.truncate(turn, direction)
				if turn in branhc and not branhc[turn]:
					del branhc[turn]
			else:
				branhc.truncate(turn, direction)

		for entities in parents.values():
			for keys in entities.values():
				for branches in keys.values():
					if branch not in branches:
						continue
					truncate_branhc(branches[branch])
		for branches in branches.values():
			if branch not in branches:
				continue
			truncate_branhc(branches[branch])
		for keys in keys.values():
			for branches in keys.values():
				if branch not in branches:
					continue
				truncate_branhc(branches[branch])
		truncate_branhc(settings[branch])
		truncate_branhc(presettings[branch])
		self.shallowest = OrderedDict()
		for entity_branch in keycache:
			if entity_branch[-1] == branch:
				truncate_branhc(keycache[entity_branch])

	@staticmethod
	def _iter_future_contradictions(entity, key, turns, branch, turn, tick,
									value):
		"""Iterate over contradicted ``(turn, tick)`` if applicable"""
		# assumes that all future entries are in the plan
		if not turns:
			return
		if turn in turns:
			future_ticks = turns[turn].future(tick)
			for tck, newval in future_ticks.items():
				if newval != value:
					yield turn, tck
			future_turns = turns.future(turn)
		elif turns.rev_gettable(turn):
			future_turns = turns.future(turn)
		else:
			future_turns = turns
		if not future_turns:
			return
		for trn, ticks in future_turns.items():
			for tick, newval in ticks.items():
				if newval != value:
					yield trn, tick

	def _store_journal(self, *args):
		# overridden in LiSE.cache.InitializedCache
		entity, key, branch, turn, tick, value = args[-6:]
		parent = args[:-6]
		settings_turns = self.settings[branch]
		presettings_turns = self.presettings[branch]
		try:
			prev = self.retrieve(*args[:-1])
		except KeyError:
			prev = None
		if turn in settings_turns or turn in settings_turns.future():
			# These assertions hold for most caches but not for the contents
			# caches, and are therefore commented out.
			# assert turn in presettings_turns \
			# or turn in presettings_turns.future()
			setticks = settings_turns[turn]
			# assert tick not in setticks
			presetticks = presettings_turns[turn]
			# assert tick not in presetticks
			presetticks[tick] = parent + (entity, key, prev)
			setticks[tick] = parent + (entity, key, value)
		else:
			presettings_turns[turn] = {tick: parent + (entity, key, prev)}
			settings_turns[turn] = {tick: parent + (entity, key, value)}

	def _base_retrieve(self, args):
		shallowest = self.shallowest
		if args in shallowest:
			return shallowest[args]
		entity = args[:-4]
		key, branch, turn, tick = args[-4:]
		keyframes = self.keyframe.get(entity, {})
		branches = self.branches
		entikey = entity + (key, )
		if entikey in branches:
			branchentk = branches[entikey]
			brancs = branchentk.get(branch)
			if brancs is not None and brancs.rev_gettable(turn):
				if turn in brancs:
					if brancs[turn].rev_gettable(tick):
						ret = brancs[turn][tick]
						shallowest[args] = ret
						return ret
					elif brancs.rev_gettable(turn - 1):
						b1 = brancs[turn - 1]
						ret = b1.final()
						shallowest[args] = ret
						return ret
				else:
					ret = brancs[turn].final()
					shallowest[args] = ret
					return ret
			for (b, r, t) in self.db._iter_parent_btt(branch):
				brancs = branchentk.get(b)
				if brancs is not None and brancs.rev_gettable(r):
					if r in brancs and brancs[r].rev_gettable(t):
						# if there's a keyframe *later* than in brancs,
						# but *earlier* than (b, r, t), use the
						# keyframe instead
						if b in keyframes and r in keyframes[b]:
							kfbr = keyframes[b][r]
							if brancs[r].rev_before(t) < kfbr.rev_before(
								t) < t:
								kf = kfbr[t]
								if key in kf:
									ret = kf[key]
									shallowest[args] = ret
									return ret
						ret = brancs[r][t]
						shallowest[args] = ret
						return ret
					elif brancs.rev_gettable(r - 1):
						if b in keyframes and keyframes[b].rev_gettable(r - 1):
							kfb = keyframes[b]
							if brancs.rev_before(r - 1) < kfb.rev_before(r -
																			1):
								kfbr = kfb[r - 1]
								kf = kfbr.final()
								if key in kf:
									ret = kf[key]
									shallowest[args] = ret
									return ret
							elif brancs.rev_before(r - 1) == kfb.rev_before(r -
																			1):
								kfbr = kfb[r - 1]
								trns = brancs[r - 1]
								if trns.end < kfbr.end:
									kf = kfbr.final()
									if key in kf:
										ret = kf[key]
										shallowest[args] = ret
										return ret
						ret = brancs[r - 1].final()
						shallowest[args] = ret
						return ret
					elif b in keyframes and r in keyframes[b] and keyframes[b][
						r].rev_gettable(t) and key in keyframes[b][r][t]:
						ret = keyframes[b][r][t][key]
						shallowest[args] = ret
						return ret
					elif b in keyframes and keyframes[b].rev_gettable(
						r - 1) and key in keyframes[b][r - 1].final():
						ret = keyframes[b][r - 1].final()[key]
						shallowest[args] = ret
						return ret
				elif b in keyframes:
					kfb = keyframes[branch]
					if r in kfb:
						kfbr = kfb[turn]
						if kfbr.rev_gettable(tick):
							kf = kfbr[tick]
							if key in kf:
								ret = kf[key]
								shallowest[args] = ret
								return ret
					if kfb.rev_gettable(r - 1):
						kfbr = kfb[r]
						kf = kfbr.final()
						if key in kf:
							ret = kf[key]
							shallowest[args] = ret
							return ret
		else:
			if branch in keyframes:
				kfb = keyframes[branch]
				if turn in kfb:
					kfbr = kfb[turn]
					if kfbr.rev_gettable(tick):
						kf = kfbr[tick]
						if key in kf:
							ret = kf[key]
							shallowest[args] = ret
							return ret
				if kfb.rev_gettable(turn - 1):
					kfbr = kfb[turn]
					kf = kfbr.final()
					if key in kf:
						ret = kf[key]
						shallowest[args] = ret
						return ret
			for (b, r, t) in self.db._iter_parent_btt(branch):
				if b in keyframes:
					kfb = keyframes[b]
					if r in kfb:
						kfbr = kfb[r]
						if kfbr.rev_gettable(t):
							kf = kfbr[t]
							if key in kf:
								ret = kf[key]
								shallowest[args] = ret
								return ret
					if kfb.rev_gettable(r - 1):
						kfbr = kfb[r]
						kf = kfbr.final()
						if key in kf:
							ret = kf[key]
							shallowest[args] = ret
							return ret
		return KeyError(f"No value for {entity} at {branch, turn, tick}")

	def retrieve(self, *args):
		"""Get a value previously .store(...)'d.

        Needs at least five arguments. The -1th is the tick
        within the turn you want,
        the -2th is that turn, the -3th is the branch,
        and the -4th is the key. All other arguments identify
        the entity that the key is in.

        """
		ret = self._base_retrieve(args)
		if ret is None:
			raise HistoricKeyError("Set, then deleted", deleted=True)
		elif isinstance(ret, Exception):
			raise ret
		return ret

	def iter_entities_or_keys(self, *args, forward=None):
		"""Iterate over the keys an entity has, if you specify an entity.

        Otherwise iterate over the entities themselves, or at any rate the
        tuple specifying which entity.

        """
		if forward is None:
			forward = self.db._forward
		entity = args[:-3]
		branch, turn, tick = args[-3:]
		if self.db._no_kc:
			yield from self._get_adds_dels(entity, branch, turn, tick)[0]
			return
		yield from self._get_keycache(entity,
										branch,
										turn,
										tick,
										forward=forward)

	iter_entities = iter_keys = iter_entity_keys = iter_entities_or_keys

	def count_entities_or_keys(self, *args, forward=None):
		"""Return the number of keys an entity has, if you specify an entity.

        Otherwise return the number of entities.

        """
		if forward is None:
			forward = self.db._forward
		entity = args[:-3]
		branch, turn, tick = args[-3:]
		if self.db._no_kc:
			return len(self._get_adds_dels(entity, branch, turn, tick)[0])
		return len(
			self._get_keycache(entity, branch, turn, tick, forward=forward))

	count_entities = count_keys = count_entity_keys = count_entities_or_keys

	def contains_entity_or_key(self, *args):
		"""Check if an entity has a key at the given time, if entity specified.

        Otherwise check if the entity exists.

        """
		retr = self._base_retrieve(args)
		return retr is not None and not isinstance(retr, Exception)

	contains_entity = contains_key = contains_entity_key = contains_entity_or_key


class NodesCache(Cache):
	"""A cache for remembering whether nodes exist at a given time."""
	__slots__ = ()

	def store(self,
				graph,
				node,
				branch,
				turn,
				tick,
				ex,
				*,
				planning=None,
				forward=None,
				loading=False,
				contra=True):
		if not ex:
			ex = None
		return super().store(graph,
								node,
								branch,
								turn,
								tick,
								ex,
								planning=planning,
								forward=forward,
								loading=loading,
								contra=contra)

	def _update_keycache(self, *args, forward):
		graph, node, branch, turn, tick, ex = args
		if not ex:
			ex = None
		super()._update_keycache(graph,
									node,
									branch,
									turn,
									tick,
									ex,
									forward=forward)

	def _iter_future_contradictions(self, entity, key, turns, branch, turn,
									tick, value):
		yield from super()._iter_future_contradictions(entity, key, turns,
														branch, turn, tick,
														value)
		yield from self.db._edges_cache._slow_iter_node_contradicted_times(
			branch, turn, tick, entity, key)


class EdgesCache(Cache):
	"""A cache for remembering whether edges exist at a given time."""
	__slots__ = ('destcache', 'origcache', 'predecessors', '_origcache_lru',
					'_destcache_lru', '_get_destcache_stuff',
					'_get_origcache_stuff', '_additional_store_stuff')

	@property
	def successors(self):
		return self.parents

	def __init__(self, db):

		def gettest(k):
			assert len(k) == 3, "Bad key: " + repr(k)

		def settest(k, v):
			assert len(k) == 3, "Bad key: {}, to be set to {}".format(k, v)

		Cache.__init__(self,
						db,
						kfkvs={
							'gettest': gettest,
							'settest': settest
						})
		self.destcache = PickyDefaultDict(SettingsTurnDict)
		self.origcache = PickyDefaultDict(SettingsTurnDict)
		self.predecessors = StructuredDefaultDict(3, TurnDict)
		self._origcache_lru = OrderedDict()
		self._destcache_lru = OrderedDict()
		self._get_destcache_stuff = (self.destcache, self._destcache_lru,
										self._get_keycachelike,
										self.successors,
										self._adds_dels_successors)
		self._get_origcache_stuff = (self.origcache, self._origcache_lru,
										self._get_keycachelike,
										self.predecessors,
										self._adds_dels_predecessors)
		self._additional_store_stuff = (self.db, self.predecessors,
										self.successors)

	def _update_keycache(self, *args, forward):
		super()._update_keycache(*args, forward=forward)
		dest, key, branch, turn, tick, value = args[-6:]
		graph, orig = args[:-6]
		# it's possible either of these might cause unnecessary iteration
		dests = self._get_destcache(graph,
									orig,
									branch,
									turn,
									tick,
									forward=forward)
		origs = self._get_origcache(graph,
									dest,
									branch,
									turn,
									tick,
									forward=forward)
		if value is None:
			dests = dests.difference((dest, ))
			origs = origs.difference((orig, ))
		else:
			dests = dests.union((dest, ))
			origs = origs.union((orig, ))
		self.destcache[graph, orig, branch][turn][tick] = dests
		self.origcache[graph, dest, branch][turn][tick] = origs

	def _slow_iter_node_contradicted_times(self, branch, turn, tick, graph,
											node):
		# slow and bad.
		retrieve = self._base_retrieve
		for items in (self.successors[graph, node].items(),
						self.predecessors[graph, node].items()):
			for dest, idxs in items:  # dest might really be orig
				for idx, branches in idxs.items():
					brnch = branches[branch]
					if turn in brnch:
						ticks = brnch[turn]
						for tck, present in ticks.future(tick).items():
							if tck > tick and present is not retrieve(
								(graph, node, dest, idx, branch, turn, tick)):
								yield turn, tck
					for trn, ticks in brnch.future(turn).items():
						for tck, present in ticks.items():
							if present is not retrieve(
								(graph, node, dest, idx, branch, turn, tick)):
								yield trn, tck

	def _adds_dels_successors(self,
								parentity,
								branch,
								turn,
								tick,
								*,
								stoptime=None,
								cache=None):
		graph, orig = parentity
		added = set()
		deleted = set()
		cache = cache or self.successors
		if (graph, orig) in cache and cache[graph, orig]:
			for dest in cache[graph, orig]:
				addidx, delidx = self._get_adds_dels((graph, orig, dest),
														branch,
														turn,
														tick,
														stoptime=stoptime)
				if addidx and not delidx:
					added.add(dest)
				elif delidx and not addidx:
					deleted.add(dest)
		if stoptime:
			return added, deleted
		kf = self.keyframe
		itparbtt = self.db._iter_parent_btt
		its = list(kf.items())
		for ks, v in its:
			assert len(ks) == 3, "Bad key in keyframe: " + repr(ks)
		for (grap, org, dest), kfg in its:  # too much iteration!
			if (grap, org) != (graph, orig):
				continue
			for branc, trn, tck in itparbtt(branch, turn, tick):
				if branc not in kfg:
					continue
				kfgb = kfg[branc]
				if trn in kfgb:
					kfgbr = kfgb[trn]
					if kfgbr.rev_gettable(tck):
						if kfgbr[tck][0] and dest not in deleted:
							added.add(dest)
						continue
				if kfgb.rev_gettable(trn):
					if kfgb[trn].final()[0] and dest not in deleted:
						added.add(dest)
		for ks in kf.keys():
			assert len(ks) == 3, "BBadd key in keyframe: " + repr(ks)
		return added, deleted

	def _adds_dels_predecessors(self,
								parentity,
								branch,
								turn,
								tick,
								*,
								stoptime=None,
								cache=None):
		graph, dest = parentity
		added = set()
		deleted = set()
		cache = cache or self.predecessors
		if cache[graph, dest]:
			for orig in cache[graph, dest]:
				addidx, delidx = self._get_adds_dels((graph, orig, dest),
														branch,
														turn,
														tick,
														stoptime=stoptime)
				if addidx and not delidx:
					added.add(orig)
				elif delidx and not addidx:
					deleted.add(orig)
		else:
			if stoptime:
				return added, deleted
			kf = self.keyframe
			itparbtt = self.db._iter_parent_btt
			for (grap, orig, dst), kfg in kf.items():  # too much iteration!
				if (grap, dst) != (graph, dest):
					continue
				for branc, trn, tck in itparbtt(branch, turn, tick):
					if branc not in kfg:
						continue
					kfgb = kfg[branc]
					if trn in kfgb:
						kfgbr = kfgb[trn]
						if kfgbr.rev_gettable(tck):
							if kfgbr[tick][0]:
								added.add(orig)
							continue
					if kfgb.rev_gettable(trn):
						if kfgb[trn].final()[0]:
							added.add(orig)
		return added, deleted

	def _get_destcache(self, graph, orig, branch, turn, tick, *, forward):
		"""Return a set of destination nodes succeeding ``orig``"""
		(destcache, destcache_lru, get_keycachelike, successors,
			adds_dels_sucpred) = self._get_destcache_stuff
		lru_append(destcache, destcache_lru,
					((graph, orig, branch), turn, tick), KEYCACHE_MAXSIZE)
		return get_keycachelike(destcache,
								successors,
								adds_dels_sucpred, (graph, orig),
								branch,
								turn,
								tick,
								forward=forward)

	def _get_origcache(self, graph, dest, branch, turn, tick, *, forward):
		"""Return a set of origin nodes leading to ``dest``"""
		(origcache, origcache_lru, get_keycachelike, predecessors,
			adds_dels_sucpred) = self._get_origcache_stuff
		lru_append(origcache, origcache_lru,
					((graph, dest, branch), turn, tick), KEYCACHE_MAXSIZE)
		return get_keycachelike(origcache,
								predecessors,
								adds_dels_sucpred, (graph, dest),
								branch,
								turn,
								tick,
								forward=forward)

	def iter_successors(self,
						graph,
						orig,
						branch,
						turn,
						tick,
						*,
						forward=None):
		"""Iterate over successors of a given origin node at a given time."""
		if self.db._no_kc:
			yield from self._adds_dels_successors((graph, orig), branch, turn,
													tick)[0]
			return
		if forward is None:
			forward = self.db._forward
		yield from self._get_destcache(graph,
										orig,
										branch,
										turn,
										tick,
										forward=forward)

	def iter_predecessors(self,
							graph,
							dest,
							branch,
							turn,
							tick,
							*,
							forward=None):
		"""Iterate over predecessors to a destination node at a given time."""
		if self.db._no_kc:
			yield from self._adds_dels_predecessors((graph, dest), branch,
													turn, tick)[0]
			return
		if forward is None:
			forward = self.db._forward
		yield from self._get_origcache(graph,
										dest,
										branch,
										turn,
										tick,
										forward=forward)

	def count_successors(self,
							graph,
							orig,
							branch,
							turn,
							tick,
							*,
							forward=None):
		"""Return the number of successors to an origin node at a given time.

        """
		if self.db._no_kc:
			return len(
				self._adds_dels_successors((graph, orig), branch, turn,
											tick)[0])
		if forward is None:
			forward = self.db._forward
		return len(
			self._get_destcache(graph,
								orig,
								branch,
								turn,
								tick,
								forward=forward))

	def count_predecessors(self,
							graph,
							dest,
							branch,
							turn,
							tick,
							*,
							forward=None):
		"""Return the number of predecessors from a destination node at a time.

        """
		if self.db._no_kc:
			return len(
				self._adds_dels_predecessors(graph, dest, branch, turn,
												tick)[0])
		if forward is None:
			forward = self.db._forward
		return len(
			self._get_origcache(graph,
								dest,
								branch,
								turn,
								tick,
								forward=forward))

	def has_successor(self,
						graph,
						orig,
						dest,
						branch,
						turn,
						tick,
						*,
						forward=None):
		"""Return whether an edge connects the origin to the destination now

        Doesn't require the edge's index, which makes it slower than retrieving
        a particular edge.

        """
		if forward is None:
			forward = self.db._forward
		return dest in self._get_destcache(graph,
											orig,
											branch,
											turn,
											tick,
											forward=forward)

	def has_predecessor(self,
						graph,
						dest,
						orig,
						branch,
						turn,
						tick,
						*,
						forward=None):
		"""Return whether an edge connects the destination to the origin now

        Doesn't require the edge's index, which makes it slower than retrieving
        a particular edge.

        """
		if forward is None:
			forward = self.db._forward
		return orig in self._get_origcache(graph,
											dest,
											branch,
											turn,
											tick,
											forward=forward)

	def store(self,
				graph,
				orig,
				dest,
				idx,
				branch,
				turn,
				tick,
				ex,
				*,
				planning=None,
				forward=None,
				loading=False,
				contra=True):
		db, predecessors, successors = self._additional_store_stuff
		if not ex:
			ex = None
		if planning is None:
			planning = db._planning
		Cache.store(self,
					graph,
					orig,
					dest,
					idx,
					branch,
					turn,
					tick,
					ex,
					planning=planning,
					forward=forward,
					loading=loading,
					contra=contra)
		predecessors[graph, dest][orig][idx][branch][turn] = successors[
			graph, orig][dest][idx][branch][turn]

	# if ex:
	#     assert self.retrieve(graph, orig, dest, idx, branch, turn, tick)
	#     assert self.has_successor(graph, orig, dest, branch, turn, tick)
	#     assert self.has_predecessor(g
	#     raph, dest, orig, branch, turn, tick)
	# else:
	#     assert self._base_retrieve(
	#     (graph, orig, dest, idx, branch, turn, tick)) in (None, KeyError)
	#     assert not self.has_successor(
	#     graph, orig, dest, branch, turn, tick)
	#     assert not self.has_predecessor(
	#     graph, dest, orig, branch, turn, tick)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The type of node that is a location.

Though both things and places are nodes, things are obliged to be
located in another node. Places are not.

"""

from .node import Node


class Place(Node):
	"""The kind of node where a thing might ultimately be located.

    LiSE entities are truthy so long as they exist, falsy if they've
    been deleted.

    """
	__slots__ = ('graph', 'db', 'node', '_rulebook', '_rulebooks',
					'_real_rule_mapping')

	extrakeys = {
		'name',
	}

	def __getitem__(self, key):
		if key == 'name':
			return self.name
		return super().__getitem__(key)

	def __repr__(self):
		return "<{}.character[{}].place[{}]>".format(repr(self.engine),
														repr(
															self['character']),
														repr(self['name']))

	def _validate_node_type(self):
		try:
			self.engine._things_cache.retrieve(self.character.name, self.name,
												*self.engine._btt())
			return False
		except:
			return True

	def delete(self):
		"""Remove myself from the world model immediately."""
		super().delete()
		self.character.place.send(self.character.place,
									key=self.name,
									val=None)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""A script to generate the SQL needed for LiSE's database backend,
and output it in JSON form.

This uses sqlalchemy to describe the queries. It extends the module of
the same name in the ``allegedb`` package. If you change anything here,
you won't be able to use your changes until you put the generated JSON
where LiSE will look for it, as in:

``python3 sqlalchemy.py >sqlite.json``

"""
from collections import OrderedDict
from functools import partial
from sqlalchemy import Table, Column, ForeignKeyConstraint, select, bindparam, \
 func, and_, or_, INT, TEXT, BOOLEAN
from sqlalchemy.sql.ddl import CreateTable, CreateIndex

BaseColumn = Column
Column = partial(BaseColumn, nullable=False)

from json import dumps

from allegedb import alchemy


def tables_for_meta(meta):
	"""Return a dictionary full of all the tables I need for LiSE. Use the
    provided metadata object.

    """
	alchemy.tables_for_meta(meta)

	# Table for global variables that are not sensitive to sim-time.
	Table('universals',
			meta,
			Column('key', TEXT, primary_key=True),
			Column('branch', TEXT, primary_key=True, default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('value', TEXT),
			sqlite_with_rowid=False)

	Table('rules',
			meta,
			Column('rule', TEXT, primary_key=True),
			sqlite_with_rowid=False)

	# Table grouping rules into lists called rulebooks.
	Table('rulebooks',
			meta,
			Column('rulebook', TEXT, primary_key=True),
			Column('branch', TEXT, primary_key=True, default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('rules', TEXT, default='[]'),
			sqlite_with_rowid=False)

	# Table for rules' triggers, those functions that return True only
	# when their rule should run (or at least check its prereqs).
	Table('rule_triggers',
			meta,
			Column('rule', TEXT, primary_key=True),
			Column('branch', TEXT, primary_key=True, default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('triggers', TEXT, default='[]'),
			ForeignKeyConstraint(('rule', ), ['rules.rule']),
			sqlite_with_rowid=False)

	# Table for rules' prereqs, functions with veto power over a rule
	# being followed
	Table('rule_prereqs',
			meta,
			Column('rule', TEXT, primary_key=True),
			Column('branch', TEXT, primary_key=True, default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('prereqs', TEXT, default='[]'),
			ForeignKeyConstraint(('rule', ), ['rules.rule']),
			sqlite_with_rowid=False)

	# Table for rules' actions, the functions that do what the rule
	# does.
	Table('rule_actions',
			meta,
			Column('rule', TEXT, primary_key=True),
			Column('branch', TEXT, primary_key=True, default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('actions', TEXT, default='[]'),
			ForeignKeyConstraint(('rule', ), ['rules.rule']),
			sqlite_with_rowid=False)

	# The top level of the LiSE world model, the character. Includes
	# rulebooks for the character itself, its units, and all the things,
	# places, and portals it contains--though those may have their own
	# rulebooks as well.

	for name in ('character_rulebook', 'unit_rulebook',
					'character_thing_rulebook', 'character_place_rulebook',
					'character_portal_rulebook'):
		Table(name,
				meta,
				Column('character', TEXT, primary_key=True),
				Column('branch', TEXT, primary_key=True, default='trunk'),
				Column('turn', INT, primary_key=True, default=0),
				Column('tick', INT, primary_key=True, default=0),
				Column('rulebook', TEXT),
				ForeignKeyConstraint(('character', ), ['graphs.graph']),
				ForeignKeyConstraint(('rulebook', ), ['rulebooks.rulebook']),
				sqlite_with_rowid=False)

	# Rules handled within the rulebook associated with one node in
	# particular.
	nrh = Table('node_rules_handled',
				meta,
				Column('character', TEXT, primary_key=True),
				Column('node', TEXT, primary_key=True),
				Column('rulebook', TEXT, primary_key=True),
				Column('rule', TEXT, primary_key=True),
				Column('branch', TEXT, primary_key=True, default='trunk'),
				Column('turn', INT, primary_key=True, default=0),
				Column('tick', INT),
				ForeignKeyConstraint(('character', 'node'),
										['nodes.graph', 'nodes.node']),
				sqlite_with_rowid=False)

	# Rules handled within the rulebook associated with one portal in
	# particular.
	porh = Table('portal_rules_handled',
					meta,
					Column('character', TEXT, primary_key=True),
					Column('orig', TEXT, primary_key=True),
					Column('dest', TEXT, primary_key=True),
					Column('rulebook', TEXT, primary_key=True),
					Column('rule', TEXT, primary_key=True),
					Column('branch', TEXT, primary_key=True, default='trunk'),
					Column('turn', INT, primary_key=True, default=0),
					Column('tick', INT),
					ForeignKeyConstraint(
						('character', 'orig', 'dest'),
						['edges.graph', 'edges.orig', 'edges.dest']),
					sqlite_with_rowid=False)

	# Table for Things, being those nodes in a Character graph that have
	# locations.
	#
	# A Thing's location can be either a Place or another Thing, as long
	# as it's in the same Character.
	Table(
		'things',
		meta,
		Column('character', TEXT, primary_key=True),
		Column('thing', TEXT, primary_key=True),
		Column('branch', TEXT, primary_key=True, default='trunk'),
		Column('turn', INT, primary_key=True, default=0),
		Column('tick', INT, primary_key=True, default=0),
		# when location is null, this node is not a thing, but a place
		Column('location', TEXT),
		ForeignKeyConstraint(('character', 'thing'),
								['nodes.graph', 'nodes.node']),
		ForeignKeyConstraint(('character', 'location'),
								['nodes.graph', 'nodes.node']),
		sqlite_with_rowid=False)

	# The rulebook followed by a given node.
	Table('node_rulebook',
			meta,
			Column('character', TEXT, primary_key=True),
			Column('node', TEXT, primary_key=True),
			Column('branch', TEXT, primary_key=True, default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('rulebook', TEXT),
			ForeignKeyConstraint(('character', 'node'),
									['nodes.graph', 'nodes.node']),
			sqlite_with_rowid=False)

	# The rulebook followed by a given Portal.
	#
	# "Portal" is LiSE's term for an edge in any of the directed
	# graphs it uses. The name is different to distinguish them from
	# Edge objects, which exist in an underlying object-relational
	# mapper called allegedb, and have a different API.
	Table('portal_rulebook',
			meta,
			Column('character', TEXT, primary_key=True),
			Column('orig', TEXT, primary_key=True),
			Column('dest', TEXT, primary_key=True),
			Column('branch', TEXT, primary_key=True, default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('rulebook', TEXT),
			ForeignKeyConstraint(
				('character', 'orig', 'dest'),
				['edges.graph', 'edges.orig', 'edges.dest']),
			sqlite_with_rowid=False)

	# The units representing one Character in another.
	#
	# In the common situation where a Character, let's say Alice has her
	# own stats and skill tree and social graph, and also has a location
	# in physical space, you can represent this by creating a Thing in
	# the Character that represents physical space, and then making that
	# Thing an unit of Alice. On its own this doesn't do anything,
	# it's just a convenient way of indicating the relation -- but if
	# you like, you can make rules that affect all units of some
	# Character, irrespective of what Character the unit is actually
	# *in*.
	Table('units',
			meta,
			Column('character_graph', TEXT, primary_key=True),
			Column('unit_graph', TEXT, primary_key=True),
			Column('unit_node', TEXT, primary_key=True),
			Column('branch', TEXT, primary_key=True, default='trunk'),
			Column('turn', INT, primary_key=True, default=0),
			Column('tick', INT, primary_key=True, default=0),
			Column('is_unit', BOOLEAN),
			ForeignKeyConstraint(('character_graph', ), ['graphs.graph']),
			ForeignKeyConstraint(('unit_graph', 'unit_node'),
									['nodes.graph', 'nodes.node']),
			sqlite_with_rowid=False)

	crh = Table(
		'character_rules_handled',
		meta,
		Column('character', TEXT, primary_key=True),
		Column('rulebook', TEXT, primary_key=True),
		Column('rule', TEXT, primary_key=True),
		Column('branch', TEXT, primary_key=True, default='trunk'),
		Column('turn', INT, primary_key=True),
		Column('tick', INT),
		ForeignKeyConstraint(
			('character', 'rulebook'),
			['character_rulebook.character', 'character_rulebook.rulebook']),
		sqlite_with_rowid=False)

	arh = Table('unit_rules_handled',
				meta,
				Column('character', TEXT, primary_key=True),
				Column('graph', TEXT, primary_key=True),
				Column('unit', TEXT, primary_key=True),
				Column('rulebook', TEXT, primary_key=True),
				Column('rule', TEXT, primary_key=True),
				Column('branch', TEXT, primary_key=True, default='trunk'),
				Column('turn', INT, primary_key=True),
				Column('tick', INT),
				ForeignKeyConstraint(
					('character', 'rulebook'),
					['unit_rulebook.character', 'unit_rulebook.rulebook']),
				sqlite_with_rowid=False)

	ctrh = Table('character_thing_rules_handled',
					meta,
					Column('character', TEXT, primary_key=True),
					Column('thing', TEXT, primary_key=True),
					Column('rulebook', TEXT, primary_key=True),
					Column('rule', TEXT, primary_key=True),
					Column('branch', TEXT, primary_key=True, default='trunk'),
					Column('turn', INT, primary_key=True),
					Column('tick', INT),
					ForeignKeyConstraint(('character', 'rulebook'), [
						'character_thing_rulebook.character',
						'character_thing_rulebook.rulebook'
					]),
					ForeignKeyConstraint(
						('character', 'thing'),
						['things.character', 'things.thing']),
					sqlite_with_rowid=False)

	cprh = Table('character_place_rules_handled',
					meta,
					Column('character', TEXT, primary_key=True),
					Column('place', TEXT, primary_key=True),
					Column('rulebook', TEXT, primary_key=True),
					Column('rule', TEXT, primary_key=True),
					Column('branch', TEXT, primary_key=True, default='trunk'),
					Column('turn', INT, primary_key=True),
					Column('tick', INT),
					ForeignKeyConstraint(('character', 'rulebook'), [
						'character_place_rulebook.character',
						'character_place_rulebook.rulebook'
					]),
					ForeignKeyConstraint(('character', 'place'),
											['nodes.graph', 'nodes.node']),
					sqlite_with_rowid=False)

	cporh = Table('character_portal_rules_handled',
					meta,
					Column('character', TEXT, primary_key=True),
					Column('orig', TEXT, primary_key=True),
					Column('dest', TEXT, primary_key=True),
					Column('rulebook', TEXT, primary_key=True),
					Column('rule', TEXT, primary_key=True),
					Column('branch', TEXT, primary_key=True, default='trunk'),
					Column('turn', INT, primary_key=True),
					Column('tick', INT),
					ForeignKeyConstraint(('character', 'rulebook'), [
						'character_portal_rulebook.character',
						'character_portal_rulebook.rulebook'
					]),
					ForeignKeyConstraint(
						('character', 'orig', 'dest'),
						['edges.graph', 'edges.orig', 'edges.dest']),
					sqlite_with_rowid=False)

	Table('turns_completed',
			meta,
			Column('branch', TEXT, primary_key=True),
			Column('turn', INT),
			sqlite_with_rowid=False)

	return meta.tables


def indices_for_table_dict(table):
	return {}


def queries(table):
	"""Given dictionaries of tables and view-queries, return a dictionary
    of all the rest of the queries I need.

    """

	def update_where(updcols, wherecols):
		"""Return an ``UPDATE`` statement that updates the columns ``updcols``
        when the ``wherecols`` match. Every column has a bound parameter of
        the same name.

        updcols are strings, wherecols are column objects

        """
		vmap = OrderedDict()
		for col in updcols:
			vmap[col] = bindparam(col)
		wheres = [c == bindparam(c.name) for c in wherecols]
		tab = wherecols[0].table
		return tab.update().values(**vmap).where(and_(*wheres))

	r = alchemy.queries_for_table_dict(table)

	rulebooks = table['rulebooks']
	r['rulebooks_update'] = update_where(['rules'], [
		rulebooks.c.rulebook, rulebooks.c.branch, rulebooks.c.turn,
		rulebooks.c.tick
	])

	for t in table.values():
		key = list(t.primary_key)
		if 'branch' in t.columns and 'turn' in t.columns and 'tick' in t.columns:
			branch = t.columns['branch']
			turn = t.columns['turn']
			tick = t.columns['tick']
			if branch in key and turn in key and tick in key:
				key = [branch, turn, tick]
		r[t.name + '_dump'] = select(list(t.c.values())).order_by(*key)
		r[t.name + '_insert'] = t.insert().values(
			tuple(bindparam(cname) for cname in t.c.keys()))
		r[t.name + '_count'] = select([func.COUNT('*')]).select_from(t)

	r['del_char_things'] = table['things'].delete().where(
		table['things'].c.character == bindparam('character'))

	r['del_char_units'] = table['units'].delete().where(
		table['units'].c.character_graph == bindparam('character'))
	things = table['things']
	r['del_things_after'] = things.delete().where(
		and_(
			things.c.character == bindparam('character'),
			things.c.thing == bindparam('thing'),
			things.c.branch == bindparam('branch'),
			or_(
				things.c.turn > bindparam('turn'),
				and_(things.c.turn == bindparam('turn'),
						things.c.tick >= bindparam('tick')))))
	units = table['units']
	r['del_units_after'] = units.delete().where(
		and_(
			units.c.character_graph == bindparam('character'),
			units.c.unit_graph == bindparam('graph'),
			units.c.unit_node == bindparam('unit'),
			units.c.branch == bindparam('branch'),
			or_(
				units.c.turn > bindparam('turn'),
				and_(units.c.turn == bindparam('turn'),
						units.c.tick >= bindparam('tick')))))
	things_to_end_clause = and_(
		things.c.character == bindparam('character'),
		things.c.branch == bindparam('branch'),
		or_(
			things.c.turn > bindparam('turn_from_a'),
			and_(things.c.turn == bindparam('turn_from_b'),
					things.c.tick >= bindparam('tick_from'))))
	r['load_things_tick_to_end'] = select(
		[things.c.thing, things.c.turn, things.c.tick,
			things.c.location]).where(things_to_end_clause)
	r['load_things_tick_to_tick'] = select(
		[things.c.thing, things.c.turn, things.c.tick,
			things.c.location]).where(
				and_(
					things_to_end_clause,
					or_(
						things.c.turn < bindparam('turn_to_a'),
						and_(things.c.turn == bindparam('turn_to_b'),
								things.c.tick <= bindparam('tick_to')))))

	for handledtab in ('character_rules_handled', 'unit_rules_handled',
						'character_thing_rules_handled',
						'character_place_rules_handled',
						'character_portal_rules_handled', 'node_rules_handled',
						'portal_rules_handled'):
		ht = table[handledtab]
		r['del_{}_turn'.format(handledtab)] = ht.delete().where(
			and_(ht.c.branch == bindparam('branch'),
					ht.c.turn == bindparam('turn')))

	branches = table['branches']

	r['branch_children'] = select(
		[branches.c.branch]).where(branches.c.parent == bindparam('branch'))

	tc = table['turns_completed']
	r['turns_completed_update'] = update_where(['turn'], [tc.c.branch])

	return r


if __name__ == '__main__':
	from sqlalchemy import MetaData
	from sqlalchemy.dialects.sqlite.pysqlite import SQLiteDialect_pysqlite

	meta = MetaData()
	r = {}
	table = tables_for_meta(meta)
	dia = SQLiteDialect_pysqlite()
	for (n, t) in table.items():
		r["create_" + n] = str(CreateTable(t).compile(dialect=dia))
		r["truncate_" + n] = str(t.delete().compile(dialect=dia))
	index = indices_for_table_dict(table)
	for (n, x) in index.items():
		r["index_" + n] = str(CreateIndex(x).compile(dialect=dia))
	query = queries(table)
	for (n, q) in query.items():
		r[n] = str(q.compile(dialect=dia))
	print(dumps(r, sort_keys=True, indent=4))
